"""accelpy Compiler module"""

import os, sys, abc, time, threading, inspect, hashlib
from numpy import ndarray
from collections import OrderedDict

from accelpy.util import Object, shellcmd, which, get_config
from accelpy.const import (lang_priority, accel_priority, vendor_priority,
                            NODEBUG)

#########################
# Generic Compilers
#########################

class Compiler(Object):
    """Compiler Base Class"""

    avails = OrderedDict()
    libext = "so"
    objext = "o"
    opt_compile_only = "-c"
    opt_debug = "-O0 -g"
    hashkey = None 

    def __init__(self, path, option=None, debug=NODEBUG):

        self.debug = debug
        self.version = []

        if isinstance(path, str):
            self.set_version(path)
            self.path = path

        elif isinstance(path, (list, tuple)):
            for p in path:
                try:
                    self.set_version(path)
                    self.path = path
                    break
                except:
                    pass

            assert self.path
        else:
            raise Exception("Unsupported compiler path type" % str(path))

        opts = self.get_option()
        self.option = option.format(default=opts) if option else opts

        if sys.platform == "darwin":
            self.libext = "dylib"

    @abc.abstractmethod
    def parse_version(self, stdout):
        pass

    @abc.abstractmethod
    def get_option(self):
        return ""

    def from_path(self, path):
        import pdb; pdb.set_trace()

    def set_version(self, path=None):
        out = shellcmd((path if path else self.path) + " " + self.opt_version)

        if out.returncode != 0:
            raise Exception("Compiler version check fails: %s" % out.stderr)

        self.version = self.parse_version(out.stdout)
        self.hashkey = hashlib.md5(out.stdout).hexdigest()[:10]

    def _compile(self, code, ext, macros):

        _macros = []
        for k, v in macros.items():
            _macros.append("-D %s=\"%s\"" % (k, str(v)) if v else ("-D "+str(k)))

        opt_macro = " ".join(_macros)
        text = (code + opt_macro + self.vendor + "".join(self.version) + ext)
        name =  hashlib.md5(text.encode("utf-8")).hexdigest()[:10]

        blddir = get_config("blddir")
        filename = name + "." + self.codeext
        codepath = os.path.join(blddir, filename)
        outfile = os.path.join(blddir, name + "." + ext)

        if not os.path.isfile(codepath):
            try:
                with open(codepath, "w") as f:
                    f.write(code)

            except FileExistsError:
                pass

        if self.debug > 0:
            debugdir = "_accelpy_debug_"

            if not os.path.isdir(debugdir):
                try:
                    os.makedirs(debugdir)

                except FileExistsError:
                    pass

            debugfile = os.path.join(debugdir, filename) 

            if not os.path.isdir(debugfile):
                try:
                    with open(debugfile, "w") as f:
                        f.write(code)

                except FileExistsError:
                    pass

        if not os.path.isfile(outfile):

            option = self.opt_debug + " " if self.debug > 0 else ""
            option += self.opt_compile_only + " " + self.get_option()

            # PGI infoopt :  -Minfo=acc -ta=tesla
            build_cmd = "{compiler} {option} {macro} -o {outfile} {infile}".format(
                            compiler=self.path, option=option, outfile=outfile,
                            infile=codepath, macro=opt_macro)

            if self.debug > 0:
                print("DEBUG: compiling: %s" % build_cmd)

            #import pdb; pdb.set_trace()

            out = shellcmd(build_cmd)
            #print(str(out.stdout).replace("\\n", "\n"))
            #print(str(out.stderr).replace("\\n", "\n"))

            if out.returncode != 0:
                errmsg = str(out.stderr).replace("\\n", "\n")
                raise Exception("Compilation fails: %s" % errmsg)

            if not os.path.isfile(outfile):
                raise Exception("Output is not generated.")

        return outfile

    def _link(self, ext, objfiles):

        objhashes = []
        for objfile in objfiles:
            objhashes.append(os.path.basename(objfile))

        text = (self.vendor + "".join(self.version) + ext + "".join(objhashes))
        name =  hashlib.md5(text.encode("utf-8")).hexdigest()[:10]

        outfile = os.path.join(get_config("blddir"), name + "." + ext)
        option = self.get_option()

        build_cmd = "{compiler} {option} -o {outfile} {objfiles}".format(
                        compiler=self.path, option=option, outfile=outfile,
                        objfiles=" ".join(objfiles))

        #print(build_cmd)
        #import pdb; pdb.set_trace()
        out = shellcmd(build_cmd)

        if out.returncode != 0:
            errmsg = str(out.stderr).replace("\\n", "\n")
            raise Exception("Compilation fails: %s" % errmsg)

        if not os.path.isfile(outfile):
            raise Exception("Output is not generated.")

        return outfile

    def compile(self, codes, macros):

        lib = None

        objfiles = []

        # build object files
        if isinstance(codes, str):
            objfiles.append(self._compile(codes, self.objext, macros))

        elif isinstance(codes, (list, tuple)):
            for code in codes:
                objfiles.append(self._compile(code, self.objext, macros))

        return self._link(self.libext, objfiles)


class CppCompiler(Compiler):

    lang = "cpp"
    codeext = "cpp"


class FortranCompiler(Compiler):

    lang = "fortran"
    codeext = "F90"


class CppCppCompiler(CppCompiler):

    accel = "cpp"


class HipCppCompiler(CppCompiler):

    accel = "hip"


class CudaCppCompiler(CppCompiler):

    accel = "cuda"
    codeext = "cu"


class OpenaccCppCompiler(CppCompiler):

    accel = "openacc_cpp"


class OpenmpCppCompiler(CppCompiler):

    accel = "openmp_cpp"


class OmptargetCppCompiler(CppCompiler):

    accel = "omptarget_cpp"


class FortranFortranCompiler(FortranCompiler):

    accel = "fortran"


class OpenaccFortranCompiler(FortranCompiler):

    accel = "openacc_fortran"


class OpenmpFortranCompiler(FortranCompiler):

    accel = "openmp_fortran"


class OmptargetFortranCompiler(FortranCompiler):

    accel = "omptarget_fortran"

################
# GNU Compilers
################

class GnuCppCompiler(CppCompiler):

    vendor = "gnu"
    opt_version = "--version"

    def __init__(self, path=None, option=None, debug=NODEBUG):

        if not path:
            path = "g++"

        super(GnuCppCompiler, self).__init__(path, option, debug=debug)

    def parse_version(self, stdout):

        items = stdout.split()
        
        if sys.platform == "darwin":
            if items[:2] == [b'Apple', b'clang']:
                return items[3].decode().split(".")
            raise Exception("Unknown version syntaxt: %s" % str(items[:3]))

        elif sys.platform == "linux":
            if items[0] == b'g++':
                return items[2].decode().split(".")
            raise Exception("Unknown version syntaxt: %s" % str(items[:2]))

        else:
            raise Exception("Platform '%s' is not supported." % str(sys.platform))

    def get_option(self):

        if sys.platform == "darwin":
            opts = "-dynamiclib -fPIC " + super(GnuCppCompiler, self).get_option()

        elif sys.platform == "linux":
            opts = "-shared -fPIC " + super(GnuCppCompiler, self).get_option()

        else:
            raise Exception("Platform '%s' is not supported." % str(sys.platform))

        return opts

class GnuCppCppCompiler(CppCppCompiler, GnuCppCompiler):
    pass


class GnuFortranCompiler(FortranCompiler):

    vendor = "gnu"
    opt_version = "--version"
    opt_moddir = "-J %s"

    def __init__(self, path=None, option=None, debug=NODEBUG):

        if not path:
            path = "gfortran"

        super(GnuFortranCompiler, self).__init__(path, option, debug=debug)

    def parse_version(self, stdout):

        items = stdout.split()
        
        if sys.platform in ("darwin", "linux"):
            if items[0] == b'GNU':
                return items[3].decode().split(".")
            raise Exception("Unknown compiler version syntax: %s" % str(items[:3]))

        else:
            raise Exception("Platform '%s' is not supported." % str(sys.platform))

    def get_option(self):

        maxline = " -ffree-line-length-none -ffixed-line-length-none"
        append_opts = self.opt_moddir % get_config("blddir") + maxline

        if sys.platform == "darwin":
            opts = ("-dynamiclib -fPIC %s " % append_opts  +
                    super(GnuFortranCompiler, self).get_option())

        elif sys.platform == "linux":
            opts = ("-shared -fPIC %s " % append_opts +
                    super(GnuFortranCompiler, self).get_option())

        else:
            raise Exception("Platform '%s' is not supported." % str(sys.platform))

        return opts


class GnuFortranFortranCompiler(FortranFortranCompiler, GnuFortranCompiler):
        pass


class GnuOpenaccCppCompiler(OpenaccCppCompiler, GnuCppCompiler):

    def get_option(self):
        return "-fopenacc " + super(GnuOpenaccCppCompiler, self).get_option()


class GnuOpenmpCppCompiler(OpenmpCppCompiler, GnuCppCompiler):

    def get_option(self):
        return "-fopenmp " + super(GnuOpenmpCppCompiler, self).get_option()


class GnuOmptargetCppCompiler(OmptargetCppCompiler, GnuCppCompiler):

    def get_option(self):
        return "-fopenmp " + super(GnuOmptargetCppCompiler, self).get_option()


class GnuOpenaccFortranCompiler(OpenaccFortranCompiler, GnuFortranCompiler):

    def get_option(self):
        return "-fopenacc " + super(GnuOpenaccFortranCompiler, self).get_option()


class GnuOpenmpFortranCompiler(OpenmpFortranCompiler, GnuFortranCompiler):

    def get_option(self):
        return "-fopenmp " + super(GnuOpenmpFortranCompiler, self).get_option()


class GnuOmptargetFortranCompiler(OmptargetFortranCompiler, GnuFortranCompiler):

    def get_option(self):
        return "-fopenmp " + super(GnuOmptargetFortranCompiler, self).get_option()


################
# Cray Compilers
################

class CrayClangCppCompiler(CppCompiler):

    vendor = "cray"
    opt_version = "--version"
    #opt_openmp = "-h omp"

    def __init__(self, path=None, option=None, debug=NODEBUG):

        if path:
            pass

        elif which("CC"):
            path = "CC"

        elif which("crayCC"):
            path = "crayCC"

        elif which("clang++"):
            path = "clang++"

        super(CrayClangCppCompiler, self).__init__(path, option, debug=debug)

    def parse_version(self, stdout):

        items = stdout.split()

        if sys.platform == "linux":
            if items[:2] == [b'Cray', b'clang']:
                return items[3].decode().split(".")
            raise Exception("Unknown version syntaxt: %s" % str(items[:3]))

        else:
            raise Exception("Platform '%s' is not supported." % str(sys.platform))

    def get_option(self):

        if sys.platform == "linux":
            opts = "-shared -fPIC " + super(CrayClangCppCompiler, self).get_option()

        else:
            raise Exception("Platform '%s' is not supported." % str(sys.platform))

        return opts


class CrayClangCppCppCompiler(CppCppCompiler, CrayClangCppCompiler):
    pass


class CrayClangOpenaccCppCompiler(OpenaccCppCompiler, CrayClangCppCompiler):

    def get_option(self):
        return "-h acc,noomp " + super(CrayClangOpenaccCppCompiler, self).get_option()


class CrayClangOpenmpCppCompiler(OpenmpCppCompiler, CrayClangCppCompiler):

    def get_option(self):
        #return "-fopenmp " + super(CrayClangOpenmpCppCompiler, self).get_option()
        return "-h omp,noacc " + super(CrayClangOpenmpCppCompiler, self).get_option()


class CrayClangOmptargetCppCompiler(OmptargetCppCompiler, CrayClangCppCompiler):

    def get_option(self):
        return "-fopenmp " + super(CrayClangOmptargetCppCompiler, self).get_option()
        #return "-h omp,noacc " + super(CrayClangOmptargetCppCompiler, self).get_option()


class CrayFortranCompiler(FortranCompiler):

    vendor = "cray"
    opt_version = "--version"
    #opt_openmp = "-h omp"
    opt_moddir = "-J %s"

    def __init__(self, path=None, option=None, debug=NODEBUG):

        if path:
            pass

        elif which("ftn"):
            path = "ftn"

        elif which("crayftn"):
            path = "crayftn"

        super(CrayFortranCompiler, self).__init__(path, option, debug=debug)

    def parse_version(self, stdout):

        items = stdout.split()
        
        if sys.platform == "linux":
            if items[:2] == [b'Cray', b'Fortran']:
                return items[4].decode().split(".")
            raise Exception("Unknown compiler version syntax: %s" % str(items[:4]))

        else:
            raise Exception("Platform '%s' is not supported." % str(sys.platform))

    def get_option(self):

        if sys.platform == "linux":
            moddir = self.opt_moddir % get_config("blddir")
            opts = ("-shared -fPIC %s " % moddir +
                    super(CrayFortranCompiler, self).get_option())
        else:
            raise Exception("Platform '%s' is not supported." % str(sys.platform))

        return opts


class CrayFortranFortranCompiler(FortranFortranCompiler, CrayFortranCompiler):
    pass


class CrayOpenaccFortranCompiler(OpenaccFortranCompiler, CrayFortranCompiler):

    def get_option(self):
        return "-h acc,noomp " + super(CrayOpenaccFortranCompiler, self).get_option()


class CrayOpenmpFortranCompiler(OpenmpFortranCompiler, CrayFortranCompiler):

    def get_option(self):
        return "-h omp,noacc " + super(CrayOpenmpFortranCompiler, self).get_option()


class CrayOmptargetFortranCompiler(OmptargetFortranCompiler, CrayFortranCompiler):

    def get_option(self):
        return "-h omp,noacc " + super(CrayOmptargetFortranCompiler, self).get_option()


################
# AMD Compilers
################

class AmdClangCppCompiler(CppCompiler):

    vendor = "amd"
    opt_version = "--version"

    def __init__(self, path=None, option=None, debug=NODEBUG):

        if path:
            pass

        elif which("amdclang++"):
            path = "amdclang++"

        super(AmdClangCppCompiler, self).__init__(path, option, debug=debug)

    def parse_version(self, stdout):

        items = stdout.split()

        if sys.platform == "linux":
            if items[:2] == [b'AMD', b'clang']:
                return items[2].decode().split(".")
            raise Exception("Unknown version syntaxt: %s" % str(items[:2]))

        else:
            raise Exception("Platform '%s' is not supported." % str(sys.platform))

    def get_option(self):

        if sys.platform == "linux":
            opts = "-shared -fPIC " + super(AmdClangCppCompiler, self).get_option()

        else:
            raise Exception("Platform '%s' is not supported." % str(sys.platform))

        return opts


class AmdClangCppCppCompiler(CppCppCompiler, AmdClangCppCompiler):
    pass


class AmdClangOpenmpCppCompiler(OpenmpCppCompiler, AmdClangCppCompiler):

    def get_option(self):
        return "-fopenmp " + super(AmdClangOpenmpCppCompiler, self).get_option()


class AmdClangOmptargetCppCompiler(OmptargetCppCompiler, AmdClangCppCompiler):

    def get_option(self):
        return "-fopenmp " + super(AmdClangOmptargetCppCompiler, self).get_option()


class AmdFlangFortranCompiler(FortranCompiler):

    vendor = "amd"
    opt_version = "--version"
    opt_moddir = "-J %s"

    def __init__(self, path=None, option=None, debug=NODEBUG):

        if path:
            pass

        elif which("amdflang"):
            path = "amdflang"

        super(AmdFlangFortranCompiler, self).__init__(path, option, debug=debug)

    def parse_version(self, stdout):

        items = stdout.split()
        
        if sys.platform == "linux":
            if items[:2] == [b'AMD', b'flang-new']:
                return items[2].decode().split(".")
            raise Exception("Unknown compiler version syntax: %s" % str(items[:2]))

        else:
            raise Exception("Platform '%s' is not supported." % str(sys.platform))

    def get_option(self):

        if sys.platform == "linux":
            moddir = self.opt_moddir % get_config("blddir")
            opts = ("-shared -fPIC %s " % moddir +
                    super(AmdFlangFortranCompiler, self).get_option())

        else:
            raise Exception("Platform '%s' is not supported." % str(sys.platform))

        return opts


class AmdFlangFortranFortranCompiler(FortranFortranCompiler, AmdFlangFortranCompiler):
    pass


class AmdFlangOpenmpFortranCompiler(OpenmpFortranCompiler, AmdFlangFortranCompiler):

    def get_option(self):
        return "-fopenmp " + super(AmdFlangOpenmpFortranCompiler, self).get_option()


class AmdFlangOpenaccFortranCompiler(OpenaccFortranCompiler, AmdFlangFortranCompiler):

    def get_option(self):
        return "-fopenacc " + super(AmdFlangOpenaccFortranCompiler, self).get_option()


class AmdFlangOmptargetFortranCompiler(OmptargetFortranCompiler, AmdFlangFortranCompiler):

    def get_option(self):
        return "-fopenmp " + super(AmdFlangOmptargetFortranCompiler, self).get_option()


class AmdHipCppCompiler(HipCppCompiler):

    vendor = "amd"
    opt_version = "--version"

    def __init__(self, path=None, option=None, debug=NODEBUG):

        if path:
            pass

        elif which("hipcc"):
            path = "hipcc"

        super(AmdHipCppCompiler, self).__init__(path, option, debug=debug)

    def parse_version(self, stdout):

        items = stdout.split()

        if sys.platform == "linux":
            if items[0] == b'HIP':
                return items[2].decode().split(".")
            raise Exception("Unknown version syntaxt: %s" % str(items[:2]))

        else:
            raise Exception("Platform '%s' is not supported." % str(sys.platform))

    def get_option(self):

        if sys.platform == "linux":
            opts = "-shared -fPIC " + super(AmdHipCppCompiler, self).get_option()

        else:
            raise Exception("Platform '%s' is not supported." % str(sys.platform))

        return opts

###################
# IBM XL Compilers
###################

class IbmXlCppCompiler(CppCompiler):

    vendor = "ibm"
    opt_version = "-qversion"

    def __init__(self, path=None, option=None, debug=NODEBUG):

        if path:
            pass

        elif which("xlc++_r"):
            path = "xlc++_r"

        elif which("xlc++"):
            path = "xlc++"

        super(IbmXlCppCompiler, self).__init__(path, option, debug=debug)

    def parse_version(self, stdout):

        items = stdout.split()

        if sys.platform == "linux":
            if items[:2] == [b'IBM', b'XL']:
                return items[5].decode().split(".")
            raise Exception("Unknown version syntaxt: %s" % str(items[:3]))

        else:
            raise Exception("'%s' platform is not supported yet." % sys.platform)

    def get_option(self):

        if sys.platform == "linux":
            opts = "-shared -fPIC " + super(IbmXlCppCompiler, self).get_option()

        else:
            raise Exception("'%s' platform is not supported yet." % sys.platform)

        return opts


class IbmXlCppCppCompiler(CppCppCompiler, IbmXlCppCompiler):
    pass


class IbmXlOpenmpCppCompiler(OpenmpCppCompiler, IbmXlCppCompiler):

    def get_option(self):
        return "-qsmp=omp " + super(IbmXlOpenmpCppCompiler, self).get_option()


class IbmXlOmptargetCppCompiler(OmptargetCppCompiler, IbmXlCppCompiler):

    def get_option(self):
        return "-qsmp=omp " + super(IbmXlOmptargetCppCompiler, self).get_option()


class IbmXlFortranCompiler(FortranCompiler):

    vendor = "ibm"
    opt_moddir = "-qmoddir=%s"
    opt_version = "-qversion"

    def __init__(self, path=None, option=None, debug=NODEBUG):

        if path:
            pass

        elif which("xlf2008_r"):
            path = "xlf2008_r"

        elif which("xlf2008"):
            path = "xlf2008"

        elif which("xlf2003_r"):
            path = "xlf2003_r"

        elif which("xlf2003"):
            path = "xlf2003"

        elif which("xlf95_r"):
            path = "xlf95_r"

        elif which("xlf95"):
            path = "xlf95"

        elif which("xlf90_r"):
            path = "xlf90_r"

        elif which("xlf90"):
            path = "xlf90"

        super(IbmXlFortranCompiler, self).__init__(path, option, debug=debug)

    def parse_version(self, stdout):

        items = stdout.split()
        
        if sys.platform == "linux":
            if items[:2] == [b'IBM', b'XL']:
                return items[5].decode().split(".")
            raise Exception("Unknown compiler version syntax: %s" % str(items[:3]))

        else:
            raise Exception("'%s' platform is not supported yet." % sys.platform)

    def get_option(self):

        if sys.platform == "linux":
            moddir = self.opt_moddir % get_config("blddir")
            opts = ("-qmkshrobj -qpic %s " % moddir +
                    super(IbmXlFortranCompiler, self).get_option())

        else:
            raise Exception("'%s' platform is not supported yet." % sys.platform)

        return opts


class IbmXlFortranFortranCompiler(FortranFortranCompiler, IbmXlFortranCompiler):
    pass


class IbmXlOpenmpFortranCompiler(OpenmpFortranCompiler, IbmXlFortranCompiler):

    def get_option(self):
        return "-qsmp=omp " + super(IbmXlOpenmpFortranCompiler, self).get_option()


class IbmXlOmptargetFortranCompiler(OmptargetFortranCompiler, IbmXlFortranCompiler):

    def get_option(self):
        return "-qsmp=omp " + super(IbmXlOmptargetFortranCompiler, self).get_option()


###################
# Nvidia Compilers
###################

class NvidiaCudaCppCompiler(CudaCppCompiler):

    vendor = "nvidia"
    opt_version = "--version"

    def __init__(self, path=None, option=None, debug=NODEBUG):

        if path:
            pass

        elif which("nvcc"):
            path = "nvcc"

        super(NvidiaCudaCppCompiler, self).__init__(path, option, debug=debug)

    def parse_version(self, stdout):

        items = stdout.split()

        if sys.platform == "linux":
            if items[0] == b'nvcc:':
                idx = items.index(b'Build')
                return items[idx+1].decode().split("_")
            raise Exception("Unknown version syntaxt: %s" % str(items[:2]))

        else:
            raise Exception("Platform '%s' is not supported." % str(sys.platform))

    def get_option(self):

        if sys.platform == "linux":
            opts = "-shared --compiler-options '-fPIC' " + super(NvidiaCudaCppCompiler, self).get_option()

        else:
            raise Exception("Platform '%s' is not supported." % str(sys.platform))

        return opts


###################
# PGI Compilers
###################

class PgiCppCompiler(CppCompiler):

    vendor = "pgi"
    opt_version = "--version"
    #opt_openmp = "-mp"

    def __init__(self, path=None, option=None, debug=NODEBUG):

        if path:
            pass

        elif which("pgc++"):
            path = "pgc++"

        super(PgiCppCompiler, self).__init__(path, option, debug=debug)

    def parse_version(self, stdout):

        items = stdout.strip().split()

        if sys.platform == "linux":
            if items[0] == b'pgc++':
                return items[1].decode().split(".")
            raise Exception("Unknown version syntaxt: %s" % str(items[:1]))

        else:
            raise Exception("'%s' platform is not supported yet." % sys.platform)


    def get_option(self):

        if sys.platform == "linux":
            opts = "-shared -fpic " + super(PgiCppCompiler, self).get_option()

        else:
            raise Exception("'%s' platform is not supported yet." % sys.platform)

        return opts


class PgiFortranCompiler(FortranCompiler):

    vendor = "pgi"
    opt_version = "--version"
    opt_moddir = "-module %s"

    def __init__(self, path=None, option=None, debug=NODEBUG):

        if path:
            pass

        elif which("pgfortran"):
            path = "pgfortran"

        super(PgiFortranCompiler, self).__init__(path, option, debug=debug)

    def parse_version(self, stdout):

        items = stdout.strip().split()

        if sys.platform == "linux":
            if items[0] == b'pgfortran':
                return items[1].decode().split(".")
            raise Exception("Unknown version syntaxt: %s" % str(items[:1]))

        else:
            raise Exception("'%s' platform is not supported yet." % sys.platform)

    def get_option(self):

        if sys.platform == "linux":
            moddir = self.opt_moddir % get_config("blddir")
            opts = ("-shared -fpic %s " % moddir +
                    super(PgiFortranCompiler, self).get_option())

        else:
            raise Exception("'%s' platform is not supported yet." % sys.platform)

        return opts


class PgiCppCppCompiler(CppCppCompiler, PgiCppCompiler):
    pass


class PgiFortranFortranCompiler(FortranFortranCompiler, PgiFortranCompiler):
    pass


class PgiOpenaccCppCompiler(OpenaccCppCompiler, PgiCppCompiler):

    def get_option(self):
        return "-acc " + super(PgiOpenaccCppCompiler, self).get_option()


class PgiOpenaccFortranCompiler(OpenaccFortranCompiler, PgiFortranCompiler):

    def get_option(self):
        return "-acc " + super(PgiOpenaccFortranCompiler, self).get_option()


class PgiOpenmpCppCompiler(OpenmpCppCompiler, PgiCppCompiler):

    def get_option(self):
        return "-mp " + super(PgiOpenmpCppCompiler, self).get_option()


class PgiOmptargetCppCompiler(OmptargetCppCompiler, PgiCppCompiler):

    def get_option(self):
        return "-mp " + super(PgiOmptargetCppCompiler, self).get_option()


class PgiOpenmpFortranCompiler(OpenmpFortranCompiler, PgiFortranCompiler):

    def get_option(self):
        return "-mp " + super(PgiOpenmpFortranCompiler, self).get_option()


class PgiOmptargetFortranCompiler(OmptargetFortranCompiler, PgiFortranCompiler):

    def get_option(self):
        return "-mp " + super(PgiOmptargetFortranCompiler, self).get_option()


###################
# Intel Compilers
###################

class IntelCppCompiler(CppCompiler):

    vendor = "intel"
    opt_version = "--version"
    #opt_openmp = "-qopenmp"

    def __init__(self, path=None, option=None, debug=NODEBUG):

        if path:
            pass

        elif which("icpc"):
            path = "icpc"

        super(IntelCppCompiler, self).__init__(path, option, debug=debug)

    def parse_version(self, stdout):

        items = stdout.strip().split()

        if sys.platform == "linux":
            if items[0] == b'icpc':
                return items[2].decode().split(".")
            raise Exception("Unknown version syntaxt: %s" % str(items[:1]))

        else:
            raise Exception("'%s' platform is not supported yet." % sys.platform)

    def get_option(self):

        if sys.platform == "linux":
            opts = "-shared -fpic " + super(IntelCppCompiler, self).get_option()

        else:
            raise Exception("'%s' platform is not supported yet." % sys.platform)

        return opts


class IntelCppCppCompiler(CppCppCompiler, IntelCppCompiler):
    pass


class IntelOpenmpCppCompiler(OpenmpCppCompiler, IntelCppCompiler):

    def get_option(self):
        return "-qopenmp " + super(IntelOpenmpCppCompiler, self).get_option()


class IntelOmptargetCppCompiler(OmptargetCppCompiler, IntelCppCompiler):

    def get_option(self):
        return "-qopenmp " + super(IntelOmptargetCppCompiler, self).get_option()


class IntelFortranCompiler(FortranCompiler):

    vendor = "intel"
    opt_version = "--version"
    #opt_openmp = "-qopenmp"
    opt_moddir = "-module %s"

    def __init__(self, path=None, option=None, debug=NODEBUG):

        if path:
            pass

        elif which("ifort"):
            path = "ifort"

        super(IntelFortranCompiler, self).__init__(path, option, debug=debug)

    def parse_version(self, stdout):

        items = stdout.strip().split()

        if sys.platform == "linux":
            if items[0] == b'ifort':
                return items[2].decode().split(".")
            raise Exception("Unknown version syntaxt: %s" % str(items[:1]))

        else:
            raise Exception("'%s' platform is not supported yet." % sys.platform)

    def get_option(self):

        if sys.platform == "linux":
            moddir = self.opt_moddir % get_config("blddir")
            return ("-shared -fpic %s " % moddir +
                    super(IntelFortranCompiler, self).get_option())
        else:
            raise Exception("'%s' platform is not supported yet." % sys.platform)

        return opts


class IntelFortranFortranCompiler(FortranFortranCompiler, IntelFortranCompiler):
    pass



class IntelOpenmpFortranCompiler(OpenmpFortranCompiler, IntelFortranCompiler):

    def get_option(self):
        return "-qopenmp " + super(IntelOpenmpFortranCompiler, self).get_option()


class IntelOmptargetFortranCompiler(OmptargetFortranCompiler, IntelFortranCompiler):

    def get_option(self):
        return "-qopenmp " + super(IntelOmptargetFortranCompiler, self).get_option()



def sort_compilers():

    def _langsort(l):
        return lang_priority.index(l.lang)

    def _accelsort(a):
        if hasattr(a, "accel"):
            return accel_priority.index(a.accel)

        else:
            return -1

    def _vendorsort(v):
        return vendor_priority.index(v.vendor)

    Compiler.avails = OrderedDict()

    for langsubc in sorted(Compiler.__subclasses__(), key=_langsort):
        lang = langsubc.lang
        Compiler.avails[lang] = OrderedDict()

        for accelsubc in sorted(langsubc.__subclasses__(), key=_accelsort):
            if not hasattr(accelsubc, "accel"):
                continue

            accel = accelsubc.accel
            Compiler.avails[lang][accel] = OrderedDict()

            for vendorsubc in sorted(accelsubc.__subclasses__(), key=_vendorsort):
                vendor = vendorsubc.vendor
                Compiler.avails[lang][accel][vendor] = vendorsubc

sort_compilers()


#CRAY	
#
#MODULES: PrgEnv-cray craype-accel-nvidia35 cudatoolkit (libsci_acc is automatically included) craype-accel-host for host cpu code 
#FORTRAN FLAGS: -h acc,noomp #OpenACC -h noacc,omp #OpenMP4.x -fpic -dynamic -lcudart -G2
#C FLAGLS: -h pragma=acc -h nopragma=omp -fpic -dynamic -lcudart -Gp 
#HELP FLAGS: -rm -h msgs
#
# 
#PGI	
#
#MODULES: PrgEnv-pgi cudatoolkit
#FORTRAN FLAGSS: -acc -ta=nvidia or -ta=multicore for host cpu code -lcudart -mcmodel=medium
#C FLAGS: -acc -ta=nvidia -lcudart -mcmodel=medium
#HELP FLAGS: -Minfo=accel

def generate_compiler(compile, debug=NODEBUG):

    clist = compile.split()

    compcls = Compiler.from_path(clist[0])
    comp = compcls(option=" ".join(clist[1:]), debug=debug)

    return comp


def get_compilers(accname, lang, compile=None, debug=NODEBUG):

    kernels = [(accname, lang)]

    compilers = []

    if compile:
        if isinstance(compile, str):
            citems = compile.split()

            if not citems:
                raise Exception("Blank compile")

            if os.path.isfile(citems[0]):
                compilers.append(generate_compiler(compile, debug=debug))

            else:
                # TODO: vendor name search
                for clang, langsubc in Compiler.avails.items():
                    for ckernel, kernelsubc in langsubc.items():
                        for cvendor, vendorsubc in kernelsubc.items():
                            if cvendor == citems[0] and clang == lang:
                                try:
                                    compilers.append(vendorsubc(
                                        option=" ".join(citems[1:]),
                                        debug=debug))
                                except:
                                    pass

        elif isinstance(compile, Compiler):
            compilers.append(compile)

        elif isinstance(compile, (list, tuple)):

            for comp in compile:
                if isinstance(comp, str):
                    citems = comp.split()

                    if not citems:
                        raise Exception("Blank compile")

                    if os.path.isfile(citems[0]):
                        try:
                            compilers.append(generate_compiler(comp, debug=debug))
                        except:
                            pass

                    else:
                        # TODO: vendor name search
                        for clang, langsubc in Compiler.avails.items():
                            for ckernel, kernelsubc in langsubc.items():
                                for cvendor, vendorsubc in kernelsubc.items():
                                    if cvendor == citems[0] and clang ==lang:
                                        try:
                                            compilers.append(vendorsubc(
                                                option=" ".join(citems[1:]),
                                                debug=debug))
                                        except:
                                            pass

                elif isinstance(comp, Compiler):
                    compilers.append(comp)

                else:
                    raise Exception("Unsupported compiler type: %s" % str(comp))
        else:
            raise Exception("Unsupported compiler type: %s" % str(compile))

    return_compilers = []
    errmsgs = []


    if compilers:
        for comp in compilers:
            if any(comp.accel==k[0] and comp.lang==k[1] for k in kernels):
                return_compilers.append(comp)

    elif compile is None:
        for acc, lang in kernels:

            if lang not in Compiler.avails:
                continue

            if acc not in Compiler.avails[lang]:
                continue

            vendors = Compiler.avails[lang][acc]

            for vendor, cls in vendors.items():
                try:
                    return_compilers.append(cls(debug=debug))
                except Exception as err:
                    errmsgs.append(str(err))


    if not return_compilers:
        if errmsgs:
            raise Exception("No compiler is found: %s" % "\n".join(errmsgs))
        else:
            raise Exception("No compiler is found: %s" % str(kernels))

    return return_compilers




"""accelpy module"""


from .util import load_sharedlib, invoke_sharedlib
from .compile import build_sharedlib
from .accel import Accel
from .kernel import Kernel
from .fortran import FortranAccel
from .cpp import CppAccel



"""accelpy utility module"""

import os, abc, ast, json, hashlib
import numpy

from ctypes import c_int32, c_int64, c_float, c_double
from subprocess import PIPE, run as subp_run
from ctypes import c_int, CDLL, RTLD_GLOBAL
from numpy.ctypeslib import ndpointer


_sysname = "generic"

c_dtypemap = {
    "int32": ["int32_t", c_int32],
    "int64": ["int64_t", c_int64],
    "float32": ["float", c_float],
    "float64": ["double", c_double]
}

f_dtypemap = {
    "int32": ["INTEGER (C_INT32_T)", c_int32],
    "int64": ["INTEGER (C_INT64_T)", c_int64],
    "float32": ["REAL (C_FLOAT)", c_float],
    "float64": ["REAL (C_DOUBLE)", c_double]
}


def get_c_dtype(arg):
    return c_dtypemap[arg["data"].dtype.name][0]


def get_f_dtype(arg):
    return f_dtypemap[arg["data"].dtype.name][0]

_cnt = 0
for k in os.environ:
    if "CRAY" in k:
        _cnt += 1
        if _cnt > 3:
            _sysname = "cray"
            break


class Object(abc.ABC):

    def __new__(cls, *vargs, **kwargs):

        obj = super(Object, cls).__new__(cls)

        return obj

class System:

    name = _sysname


_builtin_excludes = ["exec", "eval", "breakpoint", "memoryview"]

_accelpy_builtins = dict((k, v) for k, v in __builtins__.items()
                       if k not in _builtin_excludes)

_config = None

def appeval(text, env):

    if not text:
        return [], {}

    if not isinstance(text, str):
        raise Exception("Not a string")

    val = None
    lenv = {}

    stmts = ast.parse(text).body

    if len(stmts) == 1 and isinstance(stmts[-1], ast.Expr):
        val = eval(text, env, lenv)

    else:
        exec(text, env, lenv)

    return val, lenv


def funcargseval(text, lenv):

    def _p(*argv, **kw_str):
        return list(argv), kw_str

    env = dict(_accelpy_builtins)
    if isinstance(lenv, dict):
        env.update(lenv)

    env["_appeval_p"] = _p
    (vargs, kwargs), out = appeval("_appeval_p(%s)" % text, env)

    return vargs, kwargs

def shellcmd(cmd, shell=True, stdout=PIPE, stderr=PIPE,
             check=False, cwd=None):

    # return value has args, returncode, stdout, stderr
    return subp_run(cmd, shell=shell, stdout=stdout,
                    stderr=stderr, check=check, cwd=cwd)

def which(pgm):
    path=os.getenv('PATH')
    for p in path.split(os.path.pathsep):
        p=os.path.join(p,pgm)
        if os.path.exists(p) and os.access(p,os.X_OK):
            return p

def init_config(cfgdir, overwrite=False):

    libdir = os.path.join(cfgdir, "lib")
    cfgfile = os.path.join(cfgdir, "config")

    for vendor in ["cray", "amd", "nvidia", "intel", "pgi", "ibm", "gnu"]:
        vendor_path = os.path.join(libdir, vendor)
        if not os.path.isdir(vendor_path):
            try:
                os.makedirs(vendor_path)

            except FileExistsError:
                pass

    config = {
        "libdir": libdir,
        "blddir": "",
    }

    if overwrite or not os.path.isfile(cfgfile):
        with open(cfgfile, "w")  as f:
            json.dump(config, f)


def _cfgfile():

    cfgdir = os.path.join(os.path.expanduser("~"), ".accelpy")
    redirect = os.path.join(cfgdir, "redirect")

    if os.path.isfile(redirect):
        with open(redirect) as f:
            cfgdir = f.read().strip()

    return os.path.join(cfgdir, "config")


def get_config(key):

    global _config

    if _config is None:
        with open(_cfgfile()) as f:
            _config = json.load(f)

    return _config[key]


def set_config(key, value, save=False):

    # TODO: multiprocessing check

    if _config is None:
        raise Exception("config file is not initialized.")

    _config[key] = value

    if save:
        with open(_cfgfile(), "w") as f:
            json.dump(_config, f, indent=4)


def fortline_pack(items):

    lines = [""]
    maxlen = 72

    for item in items:
        if len(lines[-1]) + len(item) > maxlen:            

            lines[-1] += " &"
            lines.append("        &, %s" % item)

        elif lines[-1] == "":
            lines[-1] += item

        else:
            lines[-1] += ", " + item

    return lines


def gethash(text, length=10):

    return hashlib.md5(text.encode("utf-8")).hexdigest()[:length]


def load_sharedlib(libpath):

    return CDLL(libpath, mode=RTLD_GLOBAL)

def invoke_sharedlib(lang, libobj, funcname, *args):

    if lang == "cpp":
        flags = ("C",)

    elif lang == "fortran":
        #flags = ("F", "O", "W")
        flags = ("F",)

    func = getattr(libobj, funcname)
    func.restype = c_int
    func.argtypes = [ndpointer(a.dtype, flags=flags) for a in args]

    return func(*args)


def get_system():
    return System()

def pack_arguments(data, updateto=None, updatefrom=None, prefix=None,
                    setnames=None):

    res = []
    utoids = []
    ufromids = []

    if updateto is not None:
        utoids = [id(uto) for uto in updateto]

    if updatefrom is not None:
        ufromids = [id(ufrom) for ufrom in updatefrom]

    if data is not None:
        for idx, arg in enumerate(data):
            idarg = id(arg)

            narg = arg if isinstance(arg, numpy.ndarray) else numpy.asarray(arg)

            curname = "apyvar_" if prefix is None else prefix+str(idx)
            if setnames is not None:
                curname = setnames[idx]

            packed = {"data": narg, "id": idarg, "curname": curname, "orgdata": arg}

            packed["updateto"] = True if idarg in utoids else False
            packed["updatefrom"] = True if idarg in ufromids else False

            res.append(packed)

    return res




"""accelpy compile module"""

import os, sys

from collections import OrderedDict
from tempfile import TemporaryDirectory
from accelpy.util import shellcmd, get_system

# NOTE: the priorities of the compilers are defined by the order of the compiler definitions

builtin_compilers = OrderedDict()

builtin_compilers["nvidia_cpp_cuda"] = OrderedDict()
builtin_compilers["*_cpp_hip"] = OrderedDict()

builtin_compilers["cray_fortran_omptarget"] = OrderedDict()
builtin_compilers["cray_cpp_omptarget"] = OrderedDict()
builtin_compilers["cray_fortran_openacc"] = OrderedDict()
builtin_compilers["cray_cpp_openacc"] = OrderedDict()
builtin_compilers["cray_fortran_openmp"] = OrderedDict()
builtin_compilers["cray_cpp_openmp"] = OrderedDict()
builtin_compilers["cray_fortran_fortran"] = OrderedDict()
builtin_compilers["cray_cpp_cpp"] = OrderedDict()

builtin_compilers["amd_fortran_omptarget"] = OrderedDict()
builtin_compilers["amd_cpp_omptarget"] = OrderedDict()
builtin_compilers["amd_fortran_openacc"] = OrderedDict()
builtin_compilers["amd_cpp_openacc"] = OrderedDict()
builtin_compilers["amd_fortran_openmp"] = OrderedDict()
builtin_compilers["amd_cpp_openmp"] = OrderedDict()
builtin_compilers["amd_fortran_fortran"] = OrderedDict()
builtin_compilers["amd_cpp_cpp"] = OrderedDict()

builtin_compilers["ibm_fortran_omptarget"] = OrderedDict()
builtin_compilers["ibm_cpp_omptarget"] = OrderedDict()
builtin_compilers["ibm_fortran_openmp"] = OrderedDict()
builtin_compilers["ibm_cpp_openmp"] = OrderedDict()
builtin_compilers["ibm_fortran_fortran"] = OrderedDict()
builtin_compilers["ibm_cpp_cpp"] = OrderedDict()

builtin_compilers["intel_fortran_omptarget"] = OrderedDict()
builtin_compilers["intel_cpp_omptarget"] = OrderedDict()
builtin_compilers["intel_fortran_openmp"] = OrderedDict()
builtin_compilers["intel_cpp_openmp"] = OrderedDict()
builtin_compilers["intel_fortran_fortran"] = OrderedDict()
builtin_compilers["intel_cpp_cpp"] = OrderedDict()

builtin_compilers["pgi_fortran_openacc"] = OrderedDict()
builtin_compilers["pgi_cpp_openacc"] = OrderedDict()
builtin_compilers["pgi_fortran_openmp"] = OrderedDict()
builtin_compilers["pgi_cpp_openmp"] = OrderedDict()
builtin_compilers["pgi_fortran_fortran"] = OrderedDict()
builtin_compilers["pgi_cpp_cpp"] = OrderedDict()

builtin_compilers["gnu_fortran_omptarget"] = OrderedDict()
builtin_compilers["gnu_cpp_omptarget"] = OrderedDict()
builtin_compilers["gnu_fortran_openacc"] = OrderedDict()
builtin_compilers["gnu_cpp_openacc"] = OrderedDict()
builtin_compilers["gnu_fortran_openmp"] = OrderedDict()
builtin_compilers["gnu_cpp_openmp"] = OrderedDict()
builtin_compilers["gnu_fortran_fortran"] = OrderedDict()
builtin_compilers["gnu_cpp_cpp"] = OrderedDict()



def _gnu_version_check(check):

    if sys.platform == "darwin":
        return check.lower().lstrip().startswith(b"configured")

    else:
        return (check.lower().lstrip().startswith(b"gnu") or
            check.lower().lstrip().startswith(b"g++"))


def _pgi_version_check(check):
    return check.lower().lstrip().startswith(b"pg")

def _nvidia_cuda_check(check):
    return check.lower().lstrip().startswith(b"nvcc")

def _cray_version_check(check):
    return check.lower().lstrip().startswith(b"cray")

def _amd_version_check(check):
    return check.lower().lstrip().startswith(b"amd")

def _hip_check(check):
    return check.lower().lstrip().startswith(b"hip")

def _ibm_version_check(check):
    return check.lower().lstrip().startswith(b"ibm")

def _intel_version_check(check):
    temp = check.lower().lstrip()
    return temp.startswith(b"ifort") or temp.startswith(b"icpc")

system = get_system()

if system.name == "cray":

    builtin_compilers["cray_fortran_omptarget"]["ftnwrapper"] = {
            "check": ("ftn --version",_cray_version_check),
            "build": "ftn -shared -fPIC -fopenmp -J {moddir} -o {outpath}"
        }

    builtin_compilers["cray_cpp_omptarget"]["CCwrapper"] = {
            "check": ("CC --version",_cray_version_check),
            "build": "CC -shared -fPIC -fopenmp -o {outpath}"
        }

    builtin_compilers["cray_fortran_openacc"]["ftnwrapper"] = {
            "check": ("ftn --version",_cray_version_check),
            "build": "ftn -shared -fPIC -h acc,noomp -J {moddir} -o {outpath}"
        }

    builtin_compilers["cray_cpp_openacc"]["CCwrapper"] = {
            "check": ("CC --version",_cray_version_check),
            "build": "CC -shared -fPIC -h acc,noomp -o {outpath}"
        }

    builtin_compilers["cray_fortran_openmp"]["ftnwrapper"] = {
            "check": ("ftn --version",_cray_version_check),
            "build": "ftn -shared -fPIC -fopenmp -J {moddir} -o {outpath}"
        }

    builtin_compilers["cray_cpp_openmp"]["CCwrapper"] = {
            "check": ("CC --version",_cray_version_check),
            "build": "CC -shared -fPIC -fopenmp -o {outpath}"
        }

    builtin_compilers["cray_fortran_fortran"]["ftnwrapper"] = {
            "check": ("ftn --version",_cray_version_check),
            "build": "ftn -shared -fPIC -J {moddir} -o {outpath}"
        }

    builtin_compilers["cray_cpp_cpp"]["CCwrapper"] = {
            "check": ("CC --version",_cray_version_check),
            "build": "CC -shared -fPIC -o {outpath}"
        }

    builtin_compilers["amd_fortran_omptarget"]["ftnwrapper"] = {
            "check": ("ftn --version",_amd_version_check),
            "build": "ftn -shared -fPIC -fopenmp -J {moddir} -o {outpath}"
        }

    builtin_compilers["amd_cpp_omptarget"]["CCwrapper"] = {
            "check": ("CC --version",_amd_version_check),
            "build": "CC -shared -fPIC -fopenmp -o {outpath}"
        }

    builtin_compilers["amd_fortran_openacc"]["ftnwrapper"] = {
            "check": ("ftn --version",_amd_version_check),
            "build": "ftn -shared -fPIC -fopenacc -J {moddir} -o {outpath}"
        }

    builtin_compilers["amd_cpp_openacc"]["CCwrapper"] = {
            "check": ("CC --version",_amd_version_check),
            "build": "CC -shared -fPIC -fopenacc -o {outpath}"
        }

    builtin_compilers["amd_fortran_openmp"]["ftnwrapper"] = {
            "check": ("ftn --version",_amd_version_check),
            "build": "ftn -shared -fPIC -fopenmp -J {moddir} -o {outpath}"
        }

    builtin_compilers["amd_cpp_openmp"]["CCwrapper"] = {
            "check": ("CC --version",_amd_version_check),
            "build": "CC -shared -fPIC -fopenmp -o {outpath}"
        }

    builtin_compilers["amd_fortran_fortran"]["ftnwrapper"] = {
            "check": ("ftn --version",_amd_version_check),
            "build": "ftn -shared -fPIC -J {moddir} -o {outpath}"
        }

    builtin_compilers["amd_cpp_cpp"]["CCwrapper"] = {
            "check": ("CC --version",_amd_version_check),
            "build": "CC -shared -fPIC -o {outpath}"
        }

    builtin_compilers["intel_fortran_omptarget"]["ftnwrapper"] = {
            "check": ("ftn --version",_intel_version_check),
            "build": "ftn -shared -fpic -qopenacc -module {moddir} -o {outpath}"
        }

    builtin_compilers["intel_fortran_openmp"]["ftnwrapper"] = {
            "check": ("ftn --version",_intel_version_check),
            "build": "ftn -shared -fpic -qopenmp -module {moddir} -o {outpath}"
        }

    builtin_compilers["intel_fortran_fortran"]["ftnwrapper"] = {
            "check": ("ftn --version",_intel_version_check),
            "build": "ftn -shared -fpic -module {moddir} -o {outpath}"
        }

    builtin_compilers["gnu_fortran_omptarget"]["ftnwrapper"] = {
            "check": ("ftn --version", _gnu_version_check),
            "build": "ftn -shared -fPIC -fopenmp -J {moddir} -o {outpath} -ffree-line-length-none"
        }

    builtin_compilers["gnu_cpp_omptarget"]["ftnwrapper"] = {
            "check": ("CC --version", _gnu_version_check),
            "build": "CC -shared -fPIC -fopenmp -o {outpath}"
        }

    builtin_compilers["gnu_fortran_openmp"]["ftnwrapper"] = {
            "check": ("ftn --version", _gnu_version_check),
            "build": "ftn -shared -fPIC -fopenmp -J {moddir} -o {outpath} -ffree-line-length-none"
        }

    builtin_compilers["gnu_cpp_openmp"]["ftnwrapper"] = {
            "check": ("CC --version", _gnu_version_check),
            "build": "CC -shared -fPIC -fopenmp -o {outpath}"
        }

    builtin_compilers["gnu_fortran_fortran"]["ftnwrapper"] = {
            "check": ("ftn --version", _gnu_version_check),
            "build": "ftn -shared -fPIC -J {moddir} -o {outpath} -ffree-line-length-none"
        }

    builtin_compilers["gnu_cpp_cpp"]["ftnwrapper"] = {
            "check": ("CC --version", _gnu_version_check),
            "build": "CC -shared -fPIC -J {moddir} -o {outpath}"
        }

builtin_compilers["cray_fortran_omptarget"]["generic"] = {
        "check": ("crayftn --version",_cray_version_check),
        "build": "crayftn -shared -fPIC -h omp,noacc -J {moddir} -o {outpath}"
    }

builtin_compilers["cray_cpp_omptarget"]["generic"] = {
        "check": ("crayCC --version",_cray_version_check),
        "build": "crayCC -shared -fPIC -h omp,noacc -o {outpath}"
    }

builtin_compilers["cray_fortran_openacc"]["generic"] = {
        "check": ("crayftn --version",_cray_version_check),
        "build": "crayftn -shared -fPIC -h acc,noomp -J {moddir} -o {outpath}"
    }

builtin_compilers["cray_cpp_openacc"]["generic"] = {
        "check": ("crayCC --version",_cray_version_check),
        "build": "crayCC -shared -fPIC -h acc,noomp -o {outpath}"
    }

builtin_compilers["cray_fortran_openmp"]["generic"] = {
        "check": ("crayftn --version",_cray_version_check),
        "build": "crayftn -shared -fPIC -h omp,noacc -J {moddir} -o {outpath}"
    }

builtin_compilers["cray_cpp_openmp"]["generic"] = {
        "check": ("crayCC --version",_cray_version_check),
        "build": "crayCC -shared -fPIC -h omp,noacc -o {outpath}"
    }

builtin_compilers["cray_fortran_fortran"]["generic"] = {
        "check": ("crayftn --version",_cray_version_check),
        "build": "crayftn -shared -fPIC -J {moddir} -o {outpath}"
    }

builtin_compilers["cray_cpp_cpp"]["generic"] = {
        "check": ("crayCC --version",_cray_version_check),
        "build": "crayCC -shared -fPIC -o {outpath}"
    }

builtin_compilers["*_cpp_hip"]["generic"] = {
        "check": ("hipcc --version",_hip_check),
        "build": "hipcc -shared -fPIC -lamdhip64 -o {outpath}"
    }

builtin_compilers["amd_fortran_omptarget"]["generic"] = {
        "check": ("amdflang --version",_amd_version_check),
        "build": "amdflang -shared -fPIC -fopenmp -J {moddir} -o {outpath}"
    }

builtin_compilers["amd_cpp_omptarget"]["generic"] = {
        "check": ("amdclang++ --version",_amd_version_check),
        "build": "amdclang++ -shared -fPIC -fopenmp -o {outpath}"
    }

builtin_compilers["amd_fortran_openacc"]["generic"] = {
        "check": ("amdflang --version",_amd_version_check),
        "build": "amdflang -shared -fPIC -fopenacc -J {moddir} -o {outpath}"
    }

builtin_compilers["amd_cpp_openacc"]["generic"] = {
        "check": ("amdclang++ --version",_amd_version_check),
        "build": "amdclang++ -shared -fPIC -fopenacc -o {outpath}"
    }

builtin_compilers["amd_fortran_openmp"]["generic"] = {
        "check": ("amdflang --version",_amd_version_check),
        "build": "amdflang -shared -fPIC -fopenmp -J {moddir} -o {outpath}"
    }

builtin_compilers["amd_cpp_openmp"]["generic"] = {
        "check": ("amdclang++ --version",_amd_version_check),
        "build": "amdclang++ -shared -fPIC -fopenmp -o {outpath}"
    }

builtin_compilers["amd_fortran_fortran"]["generic"] = {
        "check": ("amdflang --version",_amd_version_check),
        "build": "amdflang -shared -fPIC -J {moddir} -o {outpath}"
    }

builtin_compilers["amd_cpp_cpp"]["generic"] = {
        "check": ("amdclang++ --version",_amd_version_check),
        "build": "amdclang++ -shared -fPIC -o {outpath}"
    }

builtin_compilers["intel_fortran_omptarget"]["generic"] = {
        "check": ("ifort --version",_intel_version_check),
        "build": "ifort -shared -fpic -qopenmp -module {moddir} -o {outpath}"
    }

builtin_compilers["intel_cpp_omptarget"]["generic"] = {
        "check": ("icpc --version",_intel_version_check),
        "build": "icpc -shared -fpic -qopenmp -o {outpath}"
    }

builtin_compilers["intel_fortran_openmp"]["generic"] = {
        "check": ("ifort --version",_intel_version_check),
        "build": "ifort -shared -fpic -qopenmp -module {moddir} -o {outpath}"
    }

builtin_compilers["intel_cpp_openmp"]["generic"] = {
        "check": ("icpc --version",_intel_version_check),
        "build": "icpc -shared -fpic -qopenmp -o {outpath}"
    }

builtin_compilers["intel_fortran_fortran"]["generic"] = {
        "check": ("ifort --version",_intel_version_check),
        "build": "ifort -shared -fpic -module {moddir} -o {outpath}"
    }

builtin_compilers["intel_cpp_cpp"]["generic"] = {
        "check": ("icpc --version",_intel_version_check),
        "build": "icpc -shared -fpic -o {outpath}"
    }

builtin_compilers["ibm_fortran_omptarget"]["generic"] = {
        "check": ("xlf_r -qversion",_ibm_version_check),
        "build": "xlf_r -qmkshrobj -qpic -qsmp=omp -qoffload -qmoddir={moddir} -o {outpath}"
    }

builtin_compilers["ibm_cpp_omptarget"]["generic"] = {
        "check": ("xlc++_r -qversion",_ibm_version_check),
        "build": "xlc++_r -qmkshrobj -qpic -qsmp=omp -qoffload -o {outpath}"
    }

builtin_compilers["ibm_fortran_openmp"]["generic"] = {
        "check": ("xlf_r -qversion",_ibm_version_check),
        "build": "xlf_r -qmkshrobj -qpic -qsmp=omp -qmoddir={moddir} -o {outpath}"
    }

builtin_compilers["ibm_cpp_openmp"]["generic"] = {
        "check": ("xlc++_r -qversion",_ibm_version_check),
        "build": "xlc++_r -qmkshrobj -qpic -qsmp=omp -o {outpath}"
    }

builtin_compilers["ibm_fortran_fortran"]["generic"] = {
        "check": ("xlf_r -qversion",_ibm_version_check),
        "build": "xlf_r -qmkshrobj -qpic -qmoddir={moddir} -o {outpath}"
    }

builtin_compilers["ibm_cpp_cpp"]["generic"] = {
        "check": ("xlc++_r -qversion",_ibm_version_check),
        "build": "xlc++_r -qmkshrobj -qpic -o {outpath}"
    }

builtin_compilers["pgi_fortran_openacc"]["generic"] = {
        "check": ("pgfortran --version", _pgi_version_check),
        "build": "pgfortran -shared -fpic -acc -module {moddir} -o {outpath}"
    }

builtin_compilers["pgi_cpp_openacc"]["generic"] = {
        "check": ("pgc++ --version", _pgi_version_check),
        "build": "pgc++ -shared -fpic -acc -o {outpath}"
    }

builtin_compilers["pgi_fortran_openmp"]["generic"] = {
        "check": ("pgfortran --version", _pgi_version_check),
        "build": "pgfortran -shared -fpic -mp -module {moddir} -o {outpath}"
    }

builtin_compilers["pgi_cpp_openmp"]["generic"] = {
        "check": ("pgc++ --version", _pgi_version_check),
        "build": "pgc++ -shared -fpic -mp -o {outpath}"
    }

builtin_compilers["pgi_fortran_fortran"]["generic"] = {
        "check": ("pgfortran --version", _pgi_version_check),
        "build": "pgfortran -shared -fpic -module {moddir} -o {outpath}"
    }

builtin_compilers["pgi_cpp_cpp"]["generic"] = {
        "check": ("pgc++ --version", _pgi_version_check),
        "build": "pgc++ -shared -fpic -o {outpath}"
    }

builtin_compilers["nvidia_cpp_cuda"]["generic"] = {
        "check": ("nvcc --version", _nvidia_cuda_check),
        "build": "nvcc -shared --compiler-options '-fPIC -lcudart' -o {outpath}"
    }

SHLIB = "-dynamiclib" if sys.platform == "darwin" else "-shared"

builtin_compilers["gnu_fortran_omptarget"]["generic"] = {
        "check": ("gfortran --version", _gnu_version_check),
        "build": "gfortran %s -fPIC -fopenmp -J {moddir} -o {outpath} -ffree-line-length-none" % SHLIB
    }

builtin_compilers["gnu_cpp_omptarget"]["generic"] = {
        "check": ("g++ --version", _gnu_version_check),
        "build": "g++ %s -fPIC -fopenmp -o {outpath}" % SHLIB
    }

builtin_compilers["gnu_fortran_openacc"]["generic"] = {
        "check": ("gfortran --version", _gnu_version_check),
        "build": "gfortran %s -fPIC -fopenacc -J {moddir} -o {outpath} -ffree-line-length-none" % SHLIB
    }

builtin_compilers["gnu_cpp_openacc"]["generic"] = {
        "check": ("g++ --version", _gnu_version_check),
        "build": "g++ %s -fPIC -fopenacc -o {outpath}" % SHLIB
    }

builtin_compilers["gnu_fortran_openmp"]["generic"] = {
        "check": ("gfortran --version", _gnu_version_check),
        "build": "gfortran %s -fPIC -fopenmp -J {moddir} -o {outpath} -ffree-line-length-none" % SHLIB
    }

builtin_compilers["gnu_cpp_openmp"]["generic"] = {
        "check": ("g++ --version", _gnu_version_check),
        "build": "g++ %s -fPIC -fopenmp -o {outpath}" % SHLIB
    }

builtin_compilers["gnu_fortran_fortran"]["generic"] = {
        "check": ("gfortran --version", _gnu_version_check),
        "build": "gfortran %s -fPIC -J {moddir} -o {outpath} -ffree-line-length-none" % SHLIB
    }

builtin_compilers["gnu_cpp_cpp"]["generic"] = {
        "check": ("g++ --version", _gnu_version_check),
        "build": "g++ %s -fPIC -o {outpath}" % SHLIB
    }


def build_sharedlib(srcfile, outfile, workdir, compile=None, opts="", vendor=None, lang=None, accel=None):

    srcpath = os.path.join(workdir, srcfile)
    outpath = os.path.join(workdir, outfile)

    if not os.path.isfile(srcpath):
        raise Exception("Source file does not exist: %s" % srcpath)

    moddir = os.path.dirname(srcpath)

    if isinstance(compile, str):
        cmd = compile.format(moddir=moddir, outpath=outpath)
        out = shellcmd(cmd + " " + srcpath)

        if out.returncode == 0:
            return outpath

        else:
            return
        
    # user or system defined compilers

    for comptype, comps in builtin_compilers.items():
        
        _vendor, _lang, _accel = comptype.split("_")

        if isinstance(vendor, (list, tuple)):
            if _vendor != "*" and _vendor not in vendor: continue

        elif isinstance(vendor, str):
            if _vendor != "*" and _vendor != vendor: continue

        if isinstance(lang, (list, tuple)):
            if _lang != "*" and _lang not in lang: continue

        elif isinstance(lang, str):
            if _lang != "*" and _lang != lang: continue

        if isinstance(accel, (list, tuple)):
            if _accel != "*" and _accel not in accel: continue

        elif isinstance(accel, str):
            if _accel != "*" and _accel != accel: continue
    
        for compid, compinfo in comps.items():

            try:
                res = shellcmd(compinfo["check"][0])
                avail = compinfo["check"][1](res.stdout)
                if avail is None:
                    avail = compinfo["check"][1](res.stderr)

                #print(avail, compid, compinfo["check"])
                if not avail: continue

                cmd = compinfo["build"].format(moddir=moddir, outpath=outpath)

                out = shellcmd("%s %s %s" % (cmd, opts, srcpath), cwd=workdir)

                #import pdb; pdb.set_trace()
                if out.returncode == 0:
                    return outpath

                print(str(out.stderr).replace("\\n", "\n"))
            except Exception as err:
                print("command fail: %s" % cmd)

    raise Exception("All build commands were failed")




"""accelpy module"""

from .kernel    import Kernel
from .accel     import AccelData
from .spec      import Spec
from .fortran   import FortranKernel, OpenmpFortranKernel, OpenaccFortranKernel
from .cpp       import CppKernel, OpenmpCppKernel, OpenaccCppKernel
from .hip       import HipKernel
from .cuda      import CudaKernel



"""accelpy CUDA and HIP Accelerator module"""

import os, sys

from collections import OrderedDict

from accelpy.util import get_c_dtype
from accelpy.accel import AccelBase


datasrc = """
#include <stdint.h>
#include <stdio.h>

{include}

{checkdef}

{moddvars}

extern "C" int64_t dataenter_{runid}({enterargs}) {{

    int64_t res;

    {entercopy}

    res = 0;

    return res;

}}

extern "C" int64_t dataexit_{runid}({exitargs}) {{

    int64_t res;

    {exitcopy}

    res = 0;

    return res;

}}
"""

kernelsrc = """
#include <stdint.h>
#include <stdio.h>

{include}

{checkdef}

{externs}

{macrodefs}

__global__ void device_kernel_{runid}({kernelargs}) {{

    {spec}

}}

extern "C" int64_t runkernel_{runid}({runkernelargs}) {{
    int64_t res;

    {kernelenter}

    {launchconf}

    device_kernel_{runid}<<<gridsize, blocksize>>>({launchargs});

    {kernelexit}

    res = 0;

    return res;
}}
"""


class CudaHipAccelBase(AccelBase):

    lang = "cpp"

    @classmethod
    def _gen_macrodefs(cls, localvars, modvars):

        typedefs = []
        consts = []
        macros = []

        macros.append("#define TYPE(varname) apy_type_##varname")
        macros.append("#define SHAPE(varname, dim) apy_shape_##varname##dim")
        macros.append("#define SIZE(varname) apy_size_##varname")

        for mhname, arg in modvars:
            dtype = get_c_dtype(arg)
            name = arg["curname"]
            ndim = arg["data"].ndim

            consts.append("const int64_t apy_size_%s = %d;" % (name, arg["data"].size))
            typedefs.append("typedef %s apy_type_%s;" % (dtype, name))

            if ndim > 0:

                shapestr = "".join(["[%d]"%s for s in arg["data"].shape])
                for d, s in enumerate(arg["data"].shape):
                    consts.append("const int64_t apy_shape_%s%d = %d;" % (name, d, s))
            else:
                pass

        for arg in localvars:
            dtype = get_c_dtype(arg)
            name = arg["curname"]
            ndim = arg["data"].ndim

            consts.append("const int64_t apy_size_%s = %d;" % (name, arg["data"].size))
            typedefs.append("typedef %s apy_type_%s;" % (dtype, name))

            if ndim > 0:

                shapestr = "".join(["[%d]"%s for s in arg["data"].shape])
                for d, s in enumerate(arg["data"].shape):
                    consts.append("const int64_t apy_shape_%s%d = %d;" % (name, d, s))
            else:
                pass

        return "\n".join(macros) + "\n\n" + "\n".join(typedefs) + "\n\n" + "\n".join(consts)

    @classmethod
    def _gen_launchconf(cls, secattr):

        grid, block = "dim3 gridsize(1);", "dim3 blocksize(1);"

        if "gridsize" in secattr:
            _grid = secattr["gridsize"]

            if isinstance(_grid , str):
                grid = _grid

            elif isinstance(_grid, int):
                grid = "dim3 gridsize(%d);" % _grid

            else:
                grid= "dim3 gridsize(%s);" % ", ".join([str(i) for i in _grid])

        if "blocksize" in secattr:
            _block = secattr["blocksize"]

            if isinstance(_block , str):
                block = _block

            elif isinstance(_block, int):
                block = "dim3 blocksize(%d);" % _block

            else:
                block= "dim3 blocksize(%s);" % ", ".join([str(i) for i in _block])

        return "%s\n%s\n" % (grid, block)

    @classmethod
    def gen_kernelfile(cls, knlhash, dmodname, runid, specid, modattr, section, workdir, localvars, modvars):

        kernelpath = os.path.join(workdir, "K%s%s" % (knlhash[2:], cls.srcext))

        attrspec = section.kwargs.get("attrspec", {})
        modattr.update(attrspec)

        externs = []
        runkernelargs = []
        kernelargs = []
        shapes = []
        launchargs = []
        kernelenter = []
        kernelexit = []

        for mhname, arg in modvars:
            mdname = "d" + mhname


            if arg["data"].ndim > 0:
                dtype = get_c_dtype(arg)
                hname = arg["curname"]
                shape = "".join(["[%d]"%s for s in arg["data"].shape])
                externs.append("extern %s (* %s)%s;" % (dtype, mdname, shape))
                runkernelargs.append("void * %s" % hname)
                launchargs.append("*%s" % mdname)
                kernelargs.append("%s %s%s" % (dtype, hname, shape))

                if arg["updateto"]:
                    kernelenter.append(cls._updateto(hname, mdname, arg["data"].size, dtype))

                if arg["updatefrom"]:
                    kernelexit.append(cls._updatefrom(hname, mdname, arg["data"].size, dtype))

        for arg in localvars:
            dtype = get_c_dtype(arg)
            hname = arg["curname"]
            dname = "d" + hname

            runkernelargs.append("void * %s" % hname)

            if arg["data"].ndim > 0:
                shape = "".join(["[%d]"%s for s in arg["data"].shape])
                launchargs.append("*%s" % dname)
                kernelargs.append("%s %s%s" % (dtype, hname, shape))

                kernelenter.append("//hipPointerAttribute_t attr;")
                kernelenter.append("//hipPointerGetAttributes(&attr, %s);" % hname)
                kernelenter.append("//printf(\"dptr, %s : %%p\\n\", %s);" % (mname, mname))
                kernelenter.append("//printf(\"attr : %p\\n\", attr.devicePointer);")

            else:
                launchargs.append("*((%s *) %s)" % (dtype, hname))
                kernelargs.append("%s %s" % (dtype, hname))

        kernelenter.append(cls._gen_enterfini())
        kernelexit.append(cls._gen_exitfini())

        #kernelenter.append("printf(\"KENTER\\n\");")
        #kernelexit.append("printf(\"KEXIT\\n\");")

        kernelparams = {
            "runid": str(runid) + str(specid),
            "externs": "\n".join(externs),
            "include": "\n".join(cls._gen_include()),
            "runkernelargs": ", ".join(runkernelargs),
            "kernelargs": ", ".join(kernelargs),
            "launchargs": ", ".join(launchargs),
            "kernelenter": "\n".join(kernelenter),
            "kernelexit": "\n".join(kernelexit),
            "launchconf": cls._gen_launchconf(section.kwargs),
            "macrodefs": cls._gen_macrodefs(localvars, modvars),
            "checkdef" : cls._gen_checkdef(),
            "spec": "\n".join(section.body),
        }

        with open(kernelpath, "w") as fkernel:
            fkernel.write(kernelsrc.format(**kernelparams))

        #import pdb; pdb.set_trace()
        return kernelpath

    @classmethod
    def gen_datafile(cls, modname, filename, runid, workdir, copyinout,
                        copyin, copyout, alloc, attr):

        datapath = os.path.join(workdir, filename)

        dataparams = {"runid": str(runid), "datamodname": modname,
                      "include": "\n".join(cls._gen_include())}

        moddvars = []
        enterargs = []
        exitargs = []
        entercopy = []
        exitcopy = []

        for cio in copyinout:
            hname = cio["curname"]
            dname = "d" + hname
            dtype = get_c_dtype(cio)

            enterargs.append("void * " + hname)
            exitargs.append("void * " + hname)

            if cio["data"].ndim > 0:
                shape = "".join(["[%d]"%s for s in cio["data"].shape])
                moddvars.append("%s (* %s)%s;" % (dtype, dname, shape))
                entercopy.append(cls._mapto(hname, dname, cio["data"].size, dtype))
                exitcopy.append(cls._mapfrom(hname, dname, cio["data"].size, dtype))

        for ci in copyin:
            hname = ci["curname"]
            dname = "d" + hname
            dtype = get_c_dtype(ci)

            enterargs.append("void * " + hname)

            if ci["data"].ndim > 0:
                shape = "".join(["[%d]"%s for s in ci["data"].shape])
                moddvars.append("%s (* %s)%s;" % (dtype, dname, shape))
                entercopy.append(cls._mapto(hname, dname, ci["data"].size, dtype))
                exitcopy.append(cls._mapdelete(dname))

        for co in copyout:
            hname = co["curname"]
            dname = "d" + hname
            dtype = get_c_dtype(co)

            enterargs.append("void * " + hname)
            exitargs.append("void * " + hname)

            if co["data"].ndim > 0:
                shape = "".join(["[%d]"%s for s in co["data"].shape])
                moddvars.append("%s (* %s)%s;" % (dtype, dname, shape))
                entercopy.append(cls._mapalloc(dname, co["data"].size, dtype))
                exitcopy.append(cls._mapfrom(hname, dname, co["data"].size, dtype))

        for al in alloc:
            hname = al["curname"]
            dname = "d" + hname
            dtype = get_c_dtype(al)

            enterargs.append("void * " + hname)

            if al["data"].ndim > 0:
                shape = "".join(["[%d]"%s for s in al["data"].shape])
                moddvars.append("%s (* %s)%s;" % (dtype, dname, shape))
                entercopy.append(cls._mapalloc(dname, al["data"].size, dtype))
                exitcopy.append(cls._mapdelete(dname))

        entercopy.append(cls._gen_enterfini())
        exitcopy.append(cls._gen_exitfini())

        dataparams["moddvars"]  = "\n".join(moddvars)
        dataparams["entercopy"] = "\n".join(entercopy)
        dataparams["exitcopy"]  = "\n".join(exitcopy)
        dataparams["enterargs"] = ", ".join(enterargs)
        dataparams["exitargs"]  = ", ".join(exitargs)
        dataparams["checkdef"]  = cls._gen_checkdef() 

        with open(datapath, "w") as fdata:
            fdata.write(datasrc.format(**dataparams))

        #import pdb; pdb.set_trace()
        return datapath


class CudaAccel(CudaHipAccelBase):
    accel = "cuda"
    srcext = ".cu"

    @classmethod
    def _updateto(cls, hname, dname, size, tname):

        fmt = ("cudaMemcpy({dname}, {hname}, {size} * sizeof({type}), cudaMemcpyHostToDevice);\n"
               "CHECK_API();\n")

        return fmt.format(hname=hname, dname=dname, size=str(size), type=tname)

    @classmethod
    def _updatefrom(cls, hname, dname, size, tname):

        fmt = ("cudaMemcpy({hname}, {dname}, {size} * sizeof({type}), cudaMemcpyDeviceToHost);\n"
               "CHECK_API();\n")

        return fmt.format(hname=hname, dname=dname, size=str(size), type=tname)

    @classmethod
    def _mapto(cls, hname, dname, size, tname):

        return (cls._mapalloc(dname, size, tname) +
                cls._updateto(hname, dname, size, tname))

    @classmethod
    def _mapfrom(cls, hname, dname, size, tname):

        return (cls._updatefrom(hname, dname, size, tname) +
                cls._mapdelete(dname))

    @classmethod
    def _mapalloc(cls, dname, size, tname):

        fmt = ("cudaMalloc((void **)&{dname}, {size} * sizeof({type}));\n"
               "CHECK_API();")

        return fmt.format(dname=dname, size=str(size), type=tname)

    @classmethod
    def _mapdelete(cls, dname):

        return "cudaFree(%s);\nCHECK_API();"  % dname

    @classmethod
    def _gen_include(cls):

        return ["#include <cuda_runtime_api.h>", "#include <cuda.h>"]

    @classmethod
    def _gen_checkdef(cls):
        return """
void CHECK_API(void)
{
    cudaError_t err = cudaGetLastError();
    if (err != cudaSuccess)
    {
        printf("Error: %s\\n", cudaGetErrorString(err));
        exit(err);
    }
}
"""

    @classmethod
    def _gen_enterfini(cls):
        return "cudaDeviceSynchronize();\nCHECK_API();"

    @classmethod
    def _gen_exitfini(cls):
        return "cudaDeviceSynchronize();\nCHECK_API();"


class HipAccel(CudaHipAccelBase):
    accel = "hip"
    srcext = ".cpp"

    @classmethod
    def _updateto(cls, hname, dname, size, tname):

        fmt = ("hipMemcpyHtoD({dname}, {hname}, {size} * sizeof({type}));\n"
               "CHECK_API();\n")

        return fmt.format(hname=hname, dname=dname, size=str(size), type=tname)

    @classmethod
    def _updatefrom(cls, hname, dname, size, tname):

        fmt = ("hipMemcpyDtoH({hname}, {dname}, {size} * sizeof({type}));\n"
               "CHECK_API();\n")

        return fmt.format(hname=hname, dname=dname, size=str(size), type=tname)

    @classmethod
    def _mapto(cls, hname, dname, size, tname):

        return (cls._mapalloc(dname, size, tname) +
                cls._updateto(hname, dname, size, tname))

    @classmethod
    def _mapfrom(cls, hname, dname, size, tname):

        return (cls._updatefrom(hname, dname, size, tname) +
                cls._mapdelete(dname))

    @classmethod
    def _mapalloc(cls, dname, size, tname):

        fmt = ("hipMalloc((void **)&{dname}, {size} * sizeof({type}));\n"
               "CHECK_API();\n")

        return fmt.format(dname=dname, size=str(size), type=tname)

    @classmethod
    def _mapdelete(cls, dname):

        return "hipFree(%s);\nCHECK_API();\n"  % dname

    @classmethod
    def _gen_include(cls):

        return ["#include <hip/hip_runtime.h>"]

    @classmethod
    def _gen_checkdef(cls):
        return """
void CHECK_API(void)
{
    hipError_t err = hipGetLastError();
    if (err != hipSuccess)
    {
        printf("Error: %s\\n", hipGetErrorString(err));
        exit(err);
    }
}
"""

    @classmethod
    def _gen_enterfini(cls):
        return "hipDeviceSynchronize();\nCHECK_API();"

    @classmethod
    def _gen_exitfini(cls):
        return "hipDeviceSynchronize();\nCHECK_API();"




"""accelpy Accel module"""

import abc, os, tempfile, shutil

from ctypes import CDLL, RTLD_GLOBAL
from numpy.ctypeslib import ndpointer
from collections import OrderedDict
from accelpy.const import (version, NOCACHE, MEMCACHE, FILECACHE, NODEBUG,
                            MINDEBUG, MAXDEBUG, NOPROF, MINPROF, MAXPROF,
                            accel_priority)
from accelpy.util import Object, gethash, get_config, set_config, pack_arguments
from accelpy.compiler import get_compilers

from accelpy.cache import slib_cache

class AccelDataBase(Object):

    avails = OrderedDict()
    offload = False

    # get map info, build & load slib, run mapping, hand over slib to kernels
    def __init__(self, *kernels, mapto=[], maptofrom=[], mapfrom=[], compile=None,
                    mapalloc=[], cache=MEMCACHE, profile=NOPROF, debug=NODEBUG):

        self.kernels = kernels
        self.debug = debug
        self.profile = profile
        self.cache = cache
        self.compile = compile

        if not self.offload:
            for kernel in self.kernels:
                kernel.set_name(self.name)

            return

        self.cachekey = None
        self.liblang = [None, None]
        self.libpath = None 
        self.bldpath = None

        self.mapto      = pack_arguments(mapto)
        self.maptofrom  = pack_arguments(maptofrom)
        self.mapalloc  = pack_arguments(mapalloc)
        self.mapfrom    = pack_arguments(mapfrom)

        keys = [os.uname().release, version, self.name, self.kernels[0].compile]

        argindex = 0

        for maptype, data in zip(("to", "tofrom", "alloc", "from"),
                (self.mapto, self.maptofrom, self.mapalloc, self.mapfrom)):

            keys.append(maptype)

            for item in data:
                keys.append(item["data"].shape)
                keys.append(item["data"].dtype)

                item["index"] = argindex
                argindex += 1

        if self.liblang[0] is None or self.cache < MEMCACHE:
            lang, libpath, bldpath = self.build_sharedlib(gethash(str(keys)))
            self.liblang[0] = None
            self.liblang[1] = lang
            self.libpath = libpath
            self.bldpath = bldpath

        lib = None

        if self.liblang[0] is None:
            if self.libpath is not None and os.path.isfile(self.libpath):
                try:
                    lib = CDLL(self.libpath, mode=RTLD_GLOBAL)
                except:
                    pass

            if (lib is None and self.bldpath is not None and
                    os.path.isfile(self.bldpath)):
                lib = CDLL(self.bldpath, mode=RTLD_GLOBAL)

        if lib is not None:
            self.liblang[0] = lib
#
#        self.mapto      = pack_arguments(mapto)
#        self.maptofrom  = pack_arguments(maptofrom)
#        self.mapalloc  = pack_arguments(mapalloc)

        argtypes = []
        argitems = self.mapto+self.maptofrom+self.mapalloc+self.mapfrom

        for item in argitems:

            if self.liblang[1] == "cpp":
                flags = ["c_contiguous"]

            elif self.liblang[1] == "fortran":
                flags = ["f_contiguous"]

            else:
                raise Exception("Unknown language: %s" % self.liblang[1])

            argtypes.append(ndpointer(item["data"].dtype, flags=",".join(flags)))

        dataenter = getattr(self.liblang[0], "dataenter")
        dataenter.argtypes = argtypes

        dataenter(*[a["data"] for a in argitems])

        for kernel in self.kernels:
            kernel.set_name(self.name)

        if (self.libpath and self.bldpath and not os.path.isfile(self.libpath)
            and os.path.isfile(self.bldpath)):
            try:
                shutil.copyfile(self.bldpath, self.libpath)

            except FileExistsError:
                pass


    def build_sharedlib(self, ckey):

        errmsgs = []

        compile = self.compile if self.compile else self.kernels[0].compile

        compilers = get_compilers(self.name, self.lang, compile=compile,
                                    debug=self.debug)
        for comp in compilers:

            cachekey = ckey + "_" + comp.hashkey

            if self.cache >= FILECACHE and cachekey in slib_cache:
                lang, basename, libext, libpath, bldpath = slib_cache[cachekey]
                self.cachekey = cachekey

                return lang, libpath, bldpath

            libdir = os.path.join(get_config("libdir"), comp.vendor, cachekey[:2])

            if not os.path.isdir(libdir):
                try:
                    os.makedirs(libdir)

                except FileExistsError:
                    pass

            basename = cachekey[2:]
            libname = basename + "." + comp.libext
            libpath = os.path.join(libdir, libname)

            if self.cache >= FILECACHE and os.path.isfile(libpath):
                self.cachekey = cachekey
                slib_cache[self.cachekey] = (comp.lang, basename, comp.libext,
                                                    libpath, None)
                return comp.lang, libpath, None

            try:
                codes, macros = self.gen_code(comp)

                if not os.path.isdir(get_config("blddir")):
                    set_config("blddir", tempfile.mkdtemp())

                bldpath = comp.compile(codes, macros)

                self.cachekey = cachekey
                slib_cache[self.cachekey] = (comp.lang, basename, comp.libext,
                                                    libpath, bldpath)
                return comp.lang, libpath, bldpath

            except Exception as err:
                errmsgs.append(str(err))

        raise Exception("\n".join(errmsgs))

    @abc.abstractmethod
    def get_dtype(self, arg):
        pass

    @abc.abstractmethod
    def gen_code(self, compiler):
        pass

    def wait(self, timeout=None):

        for kernel in self.kernels:
            kernel.wait(timeout=timeout)

        if self.offload:
            argtypes = []
            argitems = self.mapfrom

            for item in argitems:

                if self.liblang[1] == "cpp":
                    flags = ["c_contiguous"]

                elif self.liblang[1] == "fortran":
                    flags = ["f_contiguous"]

                else:
                    raise Exception("Unknown language: %s" % self.liblang[1])

                argtypes.append(ndpointer(item["data"].dtype, flags=",".join(flags)))

            dataexit = getattr(self.liblang[0], "dataexit")
            dataexit.argtypes = argtypes

            dataexit(*[a["data"] for a in argitems])

    def stop(self, timeout=None):

        self.wait(timeout=timeout)

        for kernel in self.kernels:
            kernel.stop(timeout=timeout)


def AccelData(*kernels, acctype=None, mapto=[], maptofrom=[], mapfrom=[],
                mapalloc=[], cache=MEMCACHE, profile=NOPROF, debug=NODEBUG,
                environ={}, compile=None):

    if isinstance(acctype, str):
        acctype = (acctype,)

    accel = set(AccelDataBase.avails.keys()) if acctype is None else set(acctype)

    for kernel in kernels:
        kernel.spec.eval_pysection(environ)

        if kernel.acctype and kernel.acctype not in accel:
            continue

        accel &= set(kernel.spec.list_sections(acctype=acctype))

    errmsgs = []

    for a in sorted(list(accel), key=(lambda x: accel_priority.index(x))):
        try:
            acc = AccelDataBase.avails[a](*kernels, mapto=mapto,
                            maptofrom=maptofrom, mapfrom=mapfrom,
                            mapalloc=mapalloc, cache=cache,
                            profile=profile, debug=debug, compile=compile)

            if debug > 0:
                print("DEBUG: using '%s' acceldata" % acc.name)

            return acc

        except Exception as err:
            errmsgs.append(repr(err))

    raise Exception("No acceldata is available: %s" % "\n".join(errmsgs))




"""accelpy CUDA kernel module"""

from accelpy.cpp import CppKernel


class CudaKernel(CppKernel):

    name = "cuda"
    lang = "cpp"

    def gen_kernel(self):

        kernel_sec = self.spec.get_section(self.section.kwargs["kernel"])

        return "\n".join(kernel_sec.body)

    def gen_startmain(self):

        argdefs = []

        fmt = "{dtype}(*apy_ptr_{name}){shape} = reinterpret_cast<{dtype}(*){shape}>(accelpy_var_{name});"

        for arg in self.data:
            dtype = self.get_dtype(arg)
            name = arg["curname"]
            ndim = arg["data"].ndim

            if ndim > 0:
                shape = "".join(["[%d]"%s for s in arg["data"].shape])
                argdefs.append(fmt.format(dtype=dtype, name=name, shape=shape))
                argdefs.append("%s(*apy_dptr_%s)%s;" % (dtype, name, shape))

        return "\n".join(argdefs) + "\n\n" + "\n".join(self.section.body)


CudaKernel.avails[CudaKernel.name] = CudaKernel



"""accelpy Cpp-based Accelerator module"""

import os

from collections import OrderedDict

from accelpy.util import get_c_dtype
from accelpy.accel import AccelBase


convfmt = "{dtype}(*apy_ptr_{name}){shape} = reinterpret_cast<{dtype}(*){shape}>({orgname});"

datasrc = """
#include <stdint.h>

{include}

extern "C" int64_t dataenter_{runid}({enterargs}) {{

    int64_t res;

    {enterdirective}

    res = 0;

    return res;

}}

extern "C" int64_t dataexit_{runid}({exitargs}) {{

    int64_t res;

    {exitdirective}

    res = 0;

    return res;

}}
"""

kernelsrc = """
#include <stdio.h>
#include <stdint.h>

{include}


{macrodefs}

int64_t kernel_{runid}({kernelargs}) {{
    int64_t res;

    {spec}

    res = 0;

    return res;

}}

extern "C" int64_t runkernel_{runid}({runkernelargs}) {{
    int64_t res;

    res = kernel_{runid}({actualargs});

    return res;
}}
"""


class CppAccelBase(AccelBase):

    lang = "cpp"
    srcext = ".cpp"

    @classmethod
    def _gen_include(cls):
        return []

    @classmethod
    def _gen_macrodefs(cls, localvars, modvars):

        typedefs = []
        consts = []
        macros = []

        macros.append("#define TYPE(varname) apy_type_##varname")
        macros.append("#define SHAPE(varname, dim) apy_shape_##varname##dim")
        macros.append("#define SIZE(varname) apy_size_##varname")

        for oldname, arg in modvars:
            dtype = get_c_dtype(arg)
            name = arg["curname"]
            ndim = arg["data"].ndim

            consts.append("const int64_t apy_size_%s = %d;" % (name, arg["data"].size))
            typedefs.append("typedef %s apy_type_%s;" % (dtype, name))

            if ndim > 0:

                shapestr = "".join(["[%d]"%s for s in arg["data"].shape])
                for d, s in enumerate(arg["data"].shape):
                    consts.append("const int64_t apy_shape_%s%d = %d;" % (name, d, s))
            else:
                pass

        for arg in localvars:
            dtype = get_c_dtype(arg)
            name = arg["curname"]
            ndim = arg["data"].ndim

            consts.append("const int64_t apy_size_%s = %d;" % (name, arg["data"].size))
            typedefs.append("typedef %s apy_type_%s;" % (dtype, name))

            if ndim > 0:

                shapestr = "".join(["[%d]"%s for s in arg["data"].shape])
                for d, s in enumerate(arg["data"].shape):
                    consts.append("const int64_t apy_shape_%s%d = %d;" % (name, d, s))
            else:
                pass

        return "\n".join(macros) + "\n\n" + "\n".join(typedefs) + "\n\n" + "\n".join(consts)

    @classmethod
    def gen_datafile(cls, modname, filename, runid, workdir, copyinout,
                        copyin, copyout, alloc, attr):

        datapath = os.path.join(workdir, filename)

        dataparams = {"runid": str(runid), "datamodname": modname}

        enterargs = []
        exitargs = []
        enterdirective = []
        exitdirective = []

        cionames = []
        cinames = []
        conames = []
        alnames = []

        for cio in copyinout:
            cioname = cio["curname"]
            dtype = get_c_dtype(cio)
            size = cio["data"].size

            if cio["data"].ndim > 0:
                cionames.append("(* %s)[0:%d]" % (cioname, size))
                enterargs.append("%s (* %s)[%d]" % (dtype, cioname, size))
                exitargs.append("%s (* %s)[%d]" % (dtype, cioname, size))

        if cionames:
            enterdirective.append(cls._mapto(cionames))
            exitdirective.append(cls._mapfrom(cionames))

        for ci in copyin:
            ciname = ci["curname"]
            dtype = get_c_dtype(ci)
            size = ci["data"].size

            if ci["data"].ndim > 0:
                cinames.append("(* %s)[0:%d]" % (ciname, size))
                enterargs.append("%s (* %s)[%d]" % (dtype, ciname, size))

        if cinames:
            enterdirective.append(cls._mapto(cinames))

        for co in copyout:
            coname = co["curname"]
            dtype = get_c_dtype(co)
            size = co["data"].size

            if co["data"].ndim > 0:
                conames.append("(* %s)[0:%d]" % (coname, size))
                enterargs.append("%s (* %s)[%d]" % (dtype, coname, size))
                exitargs.append("%s (* %s)[%d]" % (dtype, coname, size))

        if conames:
            alnames.extend(conames)
            exitdirective.append(cls._mapfrom(conames))

        for al in alloc:
            alname = al["curname"]
            dtype = get_c_dtype(al)
            size = al["data"].size

            if al["data"].ndim > 0:
                alnames.append("(* %s)[0:%d]" % (alname, size))
                enterargs.append("%s (* %s)[%d]" % (dtype, alname, size))

        if alnames:
            enterdirective.append(cls._mapalloc(alnames))

        dataparams["enterargs"] = ", ".join(enterargs)
        dataparams["enterdirective"] =  "\n".join(enterdirective)
        dataparams["exitargs"] = ", ".join(exitargs)
        dataparams["exitdirective"] = "\n".join(exitdirective)
        dataparams["include"] = "\n".join(cls._gen_include())

        with open(datapath, "w") as fdata:
            fdata.write(datasrc.format(**dataparams))

        #import pdb; pdb.set_trace()
        return datapath

    @classmethod
    def gen_kernelfile(cls, knlhash, dmodname, runid, specid, modattr, section, workdir, localvars, modvars):

        kernelpath = os.path.join(workdir, "K%s%s" % (knlhash[2:], cls.srcext))

        attrspec = section.kwargs.get("attrspec", {})
        modattr.update(attrspec)

        runkernelargs = []
        kernelargs = []
        actualargs = []
        casts = []

        for mname, mvar in modvars:
            ndim = mvar["data"].ndim
            dtype = get_c_dtype(mvar)
            vname = mvar["curname"]

            if ndim > 0:
                shape = "".join(["[%d]"%s for s in mvar["data"].shape])
                runkernelargs.append("%s (* %s)%s" % (dtype, vname, shape))
                kernelargs.append("%s %s%s" % (dtype, vname, shape))
                actualargs.append("*%s" % vname)

            else:
                runkernelargs.append("%s * %s" % (dtype, vname))
                kernelargs.append("%s %s" % (dtype, vname))
                actualargs.append("*%s" % (dtype, vname))

        for lvar in localvars:
            ndim = lvar["data"].ndim
            dtype = get_c_dtype(lvar)
            vname = lvar["curname"]


            if ndim > 0:
                shape = "".join(["[%d]"%s for s in lvar["data"].shape])
                runkernelargs.append("%s (* %s)%s" % (dtype, vname, shape))
                kernelargs.append("%s %s%s" % (dtype, vname, shape))
                actualargs.append("*%s" % vname)

            else:
                runkernelargs.append("%s * %s" % (dtype, vname))
                kernelargs.append("%s %s" % (dtype, vname))
                actualargs.append("*%s" % (dtype, vname))

        kernelparams = {
            "runid": str(runid) + str(specid),
            "macrodefs": cls._gen_macrodefs(localvars, modvars),
            "kernelargs": ", ".join(kernelargs),
            "runkernelargs": ", ".join(runkernelargs),
            "spec": "\n".join(section.body),
            "actualargs":", ".join(actualargs),
            "include":"\n".join(cls._gen_include()) 
        }

        with open(kernelpath, "w") as fkernel:
            fkernel.write(kernelsrc.format(**kernelparams))

        #import pdb; pdb.set_trace()
        return kernelpath

class CppAccel(CppAccelBase):
    accel = "cpp"

    @classmethod
    def _mapto(cls, names):
        return ""

    @classmethod
    def _mapfrom(cls, names):
        return ""

    @classmethod
    def _mapalloc(cls, names):
        return ""


class OpenmpCppAccel(CppAccel):
    accel = "openmp"

    @classmethod
    def _gen_include(cls):
        return ["#include <omp.h>"]


class AcctargetCppAccel(CppAccelBase):
    pass


class OmptargetCppAccel(AcctargetCppAccel):
    accel = "omptarget"

    @classmethod
    def _mapto(cls, names):
        return "#pragma omp target enter data map(to:" + ", ".join(names) + ")"

    @classmethod
    def _mapfrom(cls, names):
        return "#pragma omp target exit data map(from:" + ", ".join(names) + ")"

    @classmethod
    def _mapalloc(cls, names):
        return "#pragma omp target enter data map(alloc:" + ", ".join(names) + ")"


class OpenaccCppAccel(AcctargetCppAccel):
    accel = "openacc"

    @classmethod
    def _mapto(cls, names):
        return "#pragma acc enter data copyin(" + ", ".join(names) + ")"

    @classmethod
    def _mapfrom(cls, names):
        return "#pragma acc exit data copyout(" + ", ".join(names) + ")"

    @classmethod
    def _mapalloc(cls, names):
        return "#pragma acc enter data create(" + ", ".join(names) + ")"


_cppaccels = OrderedDict()
AccelBase.avails[CppAccelBase.lang] = _cppaccels

from .cudahip import CudaAccel, HipAccel

_cppaccels[CudaAccel.accel] = CudaAccel
_cppaccels[HipAccel.accel] = HipAccel

_cppaccels[OmptargetCppAccel.accel] = OmptargetCppAccel
_cppaccels[OpenaccCppAccel.accel] = OpenaccCppAccel
_cppaccels[OpenmpCppAccel.accel] = OpenmpCppAccel
_cppaccels[CppAccel.accel] = CppAccel




"""main entry for accelpy command-line interface"""

import os
from accelpy.util import init_config, get_config, set_config
from accelpy.const import version

def cmd_config(args):

    import json

    if args.dir:

        cfgdir = os.path.join(os.path.expanduser("~"), ".accelpy")
        redirect = os.path.join(cfgdir, "redirect")

        with open(redirect, "w") as f:
            f.write(args.dir)

        init_config(args.dir)

    elif args.libdir:
        set_config("libdir", args.libdir, save=True)

            
def cmd_cache(args):

    import shutil

    if args.clear_all:
        libdir = get_config("libdir")

        if os.path.isdir(libdir):
            for item in os.listdir(libdir):

                itempath = os.path.join(libdir, item)
                if os.path.isdir(itempath):
                    shutil.rmtree(itempath)

                else:
                    os.remove(itempath)


def main():
    import argparse

    parser = argparse.ArgumentParser(description="accelpy command-line tool")
    parser.add_argument("--version", action="version", version="accelpy "+version)
    parser.add_argument("--verbose", action="store_true", help="verbose info")

    cmds = parser.add_subparsers(title='subcommands',
                description='accelpy subcommands', help='additional help')

    p_config = cmds.add_parser('config')
    p_config.add_argument("-d", "--dir", help="set path for config files")
    p_config.add_argument("-l", "--libdir", help="set path for library cache files")
    p_config.set_defaults(func=cmd_config)

    p_cache = cmds.add_parser('cache')
    p_cache.add_argument("-a", "--clear-all", action="store_true",
                            help="clear all caches")
    p_cache.set_defaults(func=cmd_cache)

    argps = parser.parse_args()

    if hasattr(argps, "func"):
        argps.func(argps)

    else:
        parser.print_help()


    return 0



"""accelpy HIP kernel module"""

from accelpy.cpp import CppKernel


class HipKernel(CppKernel):

    name = "hip"
    lang = "cpp"

    def get_include(self):
        return "#include <hip/hip_runtime.h>"

    def gen_kernel(self):

        kernel_sec = self.spec.get_section(self.section.kwargs["kernel"])

        return "\n".join(kernel_sec.body)

    def gen_startmain(self):

        argdefs = []

        fmt = "{dtype}(*apy_ptr_{name}){shape} = reinterpret_cast<{dtype}(*){shape}>(accelpy_var_{name});"

        for arg in self.data:
            dtype = self.get_dtype(arg)
            name = arg["curname"]
            ndim = arg["data"].ndim

            if ndim > 0:
                shape = "".join(["[%d]"%s for s in arg["data"].shape])
                argdefs.append(fmt.format(dtype=dtype, name=name, shape=shape))
                argdefs.append("%s(*apy_dptr_%s)%s;" % (dtype, name, shape))

        return "\n".join(argdefs) + "\n\n" + "\n".join(self.section.body)

    def gen_preprop(self):

        typedefs = []
        consts = []
        macros = []

        # TYPE(x), SHAPE(x, 0), SIZE(x), ARG(x), DVAR(x), FLATTEN(x)

        macros.append("#define TYPE(varname) apy_type_##varname")
        macros.append("#define SHAPE(varname, dim) apy_shape_##varname##dim")
        macros.append("#define SIZE(varname) apy_size_##varname")
        macros.append("#define ARG(varname) apy_type_##varname varname apy_shapestr_##varname")
        macros.append("#define VAR(varname) (*apy_ptr_##varname)")
        macros.append("#define DVAR(varname) (*apy_dptr_##varname)")
        macros.append("#define PTR(varname) apy_ptr_##varname")
        macros.append("#define DPTR(varname) apy_dptr_##varname")
        macros.append("#define FLATTEN(varname) accelpy_var_##varname")

        for arg in self.data:
            dtype = self.get_dtype(arg)
            name = arg["curname"]
            ndim = arg["data"].ndim

            consts.append("__host__ __device__ const int apy_size_%s = %d;" % (name, arg["data"].size))
            typedefs.append("typedef %s apy_type_%s;" % (dtype, name))

            if ndim > 0:

                shapestr = "".join(["[%d]"%s for s in arg["data"].shape])
                macros.append("#define apy_shapestr_%s %s" % (name, shapestr))
                for d, s in enumerate(arg["data"].shape):
                    consts.append("__host__ __device__ const int apy_shape_%s%d = %d;" % (name, d, s))
            else:
                pass

        return "\n".join(macros) + "\n\n" + "\n".join(typedefs) + "\n\n" + "\n".join(consts)

HipKernel.avails[HipKernel.name] = HipKernel



"""Accelpy kernel module"""

import os, sys, abc, tempfile, shutil, itertools

from collections import OrderedDict

from accelpy.const import version 
from accelpy.util import (Object, load_sharedlib, invoke_sharedlib,
                            pack_arguments, shellcmd, gethash, get_config) 
from accelpy.compile import build_sharedlib, builtin_compilers

class AccelBase(Object):

    libext = ".dylib" if sys.platform == "darwin" else ".so"
    avails = OrderedDict()

    @abc.abstractmethod
    def gen_datafile(cls, modname, filename, runid, workdir, copyinout,
                        copyin, copyout, alloc, attr):
        pass

    @abc.abstractmethod
    def gen_kernelfile(cls, datamodname, runid, section, workdir,
                        localvars, modvars):
        pass


class Accel:

    _ids = itertools.count(0)

    def __init__(self, copyinout=None, copyin=None, copyout=None,
                    alloc=None, compile=None, lang=None, vendor=None,
                    accel=None, attr={}, recompile=False, _debug=False):

        self._id = next(self._ids)
        self._debug = _debug
        self._lang = None
        self._accel = None
        self._attr = attr
        self._tasks = {}
        self._workdir = tempfile.mkdtemp()
        self.debug("creating workdir: %s" % self._workdir)

        self.copyinout = pack_arguments(copyinout, prefix="cio%d_" % self._id)
        self.copyin = pack_arguments(copyin, prefix="ci%d_" % self._id)
        self.copyout = pack_arguments(copyout, prefix="co%d_" % self._id)
        self.alloc = pack_arguments(alloc, prefix="al%d_" % self._id)
            
        self._libkernel = None

        keys = [os.uname().release, version, self._id]

        for data in (self.copyinout, self.copyin, self.copyout, self.alloc):
            for item in data:
                keys.append(item["data"].shape)
                keys.append(item["data"].dtype)

        orghash = gethash(str(keys))

        # user or system defined compilers
        for comptype, comps in builtin_compilers.items():
            
            _vendor, _lang, _accel = comptype.split("_")

            if isinstance(vendor, (list, tuple)):
                if _vendor != "*" and _vendor not in vendor: continue

            elif isinstance(vendor, str):
                if _vendor != "*" and _vendor != vendor: continue

            if isinstance(lang, (list, tuple)):
                if _lang != "*" and _lang not in lang: continue

            elif isinstance(lang, str):
                if _lang != "*" and _lang != lang: continue

            if isinstance(accel, (list, tuple)):
                if _accel != "*" and _accel not in accel: continue

            elif isinstance(accel, str):
                if _accel != "*" and _accel != accel: continue

            libdir = get_config("libdir")

            self._dsrchash = gethash(orghash + str(comptype) + str(_lang) + str(_accel))
            
            dfname = "D" + self._dsrchash[2:] + AccelBase.avails[_lang][_accel].srcext
            self._dfdir = os.path.join(libdir, self._dsrchash[:2])
            dsrcpath = os.path.join(self._dfdir, dfname)
            self._dmodname = "mod" + self._dsrchash[2:]

            if os.path.isfile(dsrcpath) and not recompile: 
                srcdata = dsrcpath

            else:
                srcdata = AccelBase.avails[_lang][_accel].gen_datafile(
                        self._dmodname, dfname, self._id, self._workdir,
                        self.copyinout, self.copyin, self.copyout, self.alloc,
                        self._attr)

            if srcdata is None: continue

            self.debug("data src from: %s" % srcdata)

            for compid, compinfo in comps.items():
                try:
                    res = shellcmd(compinfo["check"][0])
                    avail = compinfo["check"][1](res.stdout)
                    if avail is None:
                        avail = compinfo["check"][1](res.stderr)

                    if not avail: continue

                    cmdline = compinfo["build"]

                    self._dlibhash = gethash(self._dsrchash +str(compid) + str(cmdline))
                    dlname = "D" + self._dlibhash[2:] + AccelBase.avails[_lang][_accel].libext
                    dldir = os.path.join(libdir, self._dlibhash[:2])
                    dlibpath = os.path.join(dldir, dlname)

                    if os.path.isfile(dlibpath) and not recompile: 
                        self._libdata = self._load_run_enter(dlibpath, _lang)
                        self.debug("data lib from lib: %s" % dlibpath)

                    else:
                        dstpath = os.path.join(self._workdir, dlname)
                        modname, self._libdata = self._build_run_enter(dstpath, _lang,
                                            srcdata, cmdline)
                        self.debug("data lib from build: %s" % dstpath)

                        if not os.path.isfile(dsrcpath):
                            if not os.path.isdir(self._dfdir):
                                os.makedirs(self._dfdir)

                            shutil.copy(srcdata, dsrcpath)

                            if modname:
                                shutil.copy(os.path.join(self._workdir,
                                        modname), self._dfdir)

                        if not os.path.isfile(dlibpath):
                            if not os.path.isdir(dldir):
                                os.makedirs(dldir)

                            shutil.copy(dstpath, dlibpath)

                    self._compile = cmdline
                    self._lang = _lang
                    self._accel = _accel

                    return 

                except Exception as err:
                    print("INFO: unsuccessful compiler command: %s" % str(err))

        raise Exception("All build commands for enterdata were failed")

    def __del__(self):

        if os.path.isdir(self._workdir):
            self.debug("removing workdir: %s" % self._workdir)
            shutil.rmtree(self._workdir)

    def debug(self, *objs):

        if self._debug:
            print("DEBUG: " + " ".join([str(o) for o in objs]))

    def stop(self):

        # invoke exit function in acceldata
        exitargs = []
        exitargs.extend([cio["data"] for cio in self.copyinout])
        exitargs.extend([co["data"] for co in self.copyout])

        self.debug("libdata exit sharedlib", self._libdata)
        resdata = invoke_sharedlib(self._lang, self._libdata, "dataexit_%d" % self._id, *exitargs)

        self.debug("after dataexit cio", *[cio["data"] for cio in self.copyinout])
        self.debug("after dataexit co", *[co["data"] for co in self.copyout])

        assert resdata == 0, "dataexit invoke fail"

    def launch(self, spec, *kargs, updateto=None, updatefrom=None, macro={}, environ={}):

        localvars = pack_arguments(kargs, updateto=updateto, updatefrom=updatefrom)

        self.spec = spec
        self.spec.eval_pysection(environ)
        self.spec.update_argnames(localvars)
        self.section = self.spec.get_section(self._accel, self._lang, environ)

        if (self.section is None or self._lang not in AccelBase.avails or
            self._accel not in AccelBase.avails[self._lang]):
            raise Exception("Kernel can not be created.")

        self.section.update_argnames(localvars)

        keys = [os.uname().release, version, self._dlibhash]

        # TODO: key gen from environ and Accel env

        dids = {}
        _kargs = []
        _uonly = []
        _mvars = []

        dids = dict((d["id"], d["curname"]) for d in
                        self.copyinout+self.copyin+self.copyout+self.alloc)

        for lvar in localvars:

            if lvar["id"] in dids:
                _mvars.append(lvar)
                _uonly.append((dids[lvar["id"]], lvar))
                keys.append((dids[lvar["id"]], lvar["curname"]))

            else:
                _kargs.append(lvar)
                keys.append(lvar["data"].shape)
                keys.append(lvar["data"].dtype)

            keys.append(lvar["updateto"])
            keys.append(lvar["updatefrom"])

        self._knlhash = gethash(self.section.hash() + str(keys) + str(macro) +
                            str(environ))

        libdir = get_config("libdir")

        klib = None

        klname = "K" + self._knlhash[2:] + AccelBase.avails[
                        self._lang][self._accel].libext
        kldir = os.path.join(libdir, self._knlhash[:2])
        klibpath = os.path.join(kldir, klname)

        if self._knlhash in self._tasks:
            self._run_kernel(self._tasks[self._knlhash], _kargs, _mvars)
            self.debug("kernel run using memory cache")

        elif os.path.isfile(klibpath): 
            klib = self._load_run_kernel(klibpath, _kargs, _mvars)
            self._tasks[self._knlhash] = klib
            self.debug("kernel run using %s" % klibpath)

        else:

            self.macro = macro
            modattr = dict(self._attr)

            srckernel = AccelBase.avails[self._lang][self._accel
                                ].gen_kernelfile(self._knlhash, self._dmodname,
                                self._id, self.spec._id, modattr, self.section, self._workdir, _kargs,
                                _uonly)

            dstpath, klib = self._build_load_run_kernel(srckernel, _kargs, _mvars)

            if not os.path.isfile(klibpath):
                if not os.path.isdir(kldir):
                    os.makedirs(kldir)

                shutil.copy(dstpath, klibpath)
            self.debug("kernel run using %s" % dstpath)

        if klib and self._knlhash not in self._tasks:
            self._tasks[self._knlhash] = klib

    def _load_run_enter(self, dstdata, lang):

        # load acceldata
        libdata = load_sharedlib(dstdata)
        self.debug("libdata sharedlib", libdata)
        assert libdata is not None, "libdata load fail"

        # invoke function in acceldata
        enterargs = []
        enterargs.extend([cio["data"] for cio in self.copyinout])
        enterargs.extend([ci["data"] for ci in self.copyin])
        enterargs.extend([co["data"] for co in self.copyout])
        enterargs.extend([al["data"] for al in self.alloc])

        #import pdb;pdb.set_trace()
        resdata = invoke_sharedlib(lang, libdata, "dataenter_%d" % self._id, *enterargs)
        assert resdata == 0, "dataenter invoke fail"

        return libdata

    def _build_run_enter(self, dstdata, lang, srcdata, command):

        # build acceldata
        cmd = command.format(moddir=self._workdir, outpath=dstdata)
        out = shellcmd(cmd + " " + srcdata, cwd=self._workdir)
        #print(str(out.stdout).replace("\\n", "\n"))
        #print(str(out.stderr).replace("\\n", "\n"))
        #import pdb; pdb.set_trace()
        assert os.path.isfile(dstdata), str(out.stderr)

        modname = None

        for fname in os.listdir(self._workdir):
            if fname.lower().endswith(".mod"):
                modname = fname
                break

        return modname, self._load_run_enter(dstdata, lang)

    def _build_load_run_kernel(self, srckernel, localvars, modvars):

        libext = AccelBase.avails[self._lang][self._accel].libext

        macros = []
        for m, d in self.macro.items():
            if d is None or d is False:
                continue

            if d is True:
                d = 1

            macros.append("-D%s=%s" % (str(m), str(d)))
        
        include = "-I %s" % self._dfdir

        # build kernel
        dstkernel = os.path.splitext(srckernel)[0] + libext
        cmd = self._compile.format(moddir=self._workdir, outpath=dstkernel)
        out = shellcmd("%s %s %s %s" % (cmd, " ".join(macros), include,
                        srckernel), cwd=self._workdir)
        #print(str(out.stdout).replace("\\n", "\n"))
        #print(str(out.stderr).replace("\\n", "\n"))
        #import pdb; pdb.set_trace()
        assert os.path.isfile(dstkernel), str(out.stderr).replace("\\n", "\n")

        return dstkernel, self._load_run_kernel(dstkernel, localvars, modvars)

    def _load_run_kernel(self, libpath, localvars, modvars):

        # load accelkernel
        libkernel = load_sharedlib(libpath)
        self.debug("libkernel sharedlib", libkernel)
        assert libkernel is not None, "libkernel load fail"

        self._run_kernel(libkernel, localvars, modvars)

        return libkernel


    def _run_kernel(self, libkernel, localvars, modvars):

        # invoke function in accelkernel
        kernelargs = [mvar["data"] for mvar in modvars]
        kernelargs.extend([lvar["data"] for lvar in localvars])

        self.debug("before kernel", *kernelargs)
        #import pdb; pdb.set_trace()
        reskernel = invoke_sharedlib(self._lang, libkernel,
                        "runkernel_%d%d" % (self._id, self.spec._id),  *kernelargs)

        self.debug("after kernel cio", *kernelargs)

        assert reskernel == 0, "runkernel invoke fail"




"""accelpy C++ kernel module"""

import abc

from accelpy.accel import AccelDataBase
from accelpy.kernel import KernelBase
from accelpy.util import c_dtypemap, getname_varmap

##########################
#  Code templates
##########################

t_main = """
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <stdint.h>
{include}

{preprop}

{varmap}

{kernel}

extern "C" int64_t accelpy_start() {{

    int64_t res = 0;

    {startmain}

    return res;
}}

extern "C" int64_t accelpy_stop() {{

    int64_t res;

    res = 0;

    return res;
}}
"""

t_kernel = """
extern "C" int64_t accelpy_kernel({kernelargs}){{

    int64_t res;

    {shape}

    {spec}

    res = 0;

    return res;
}}
"""

t_varmap = """
extern "C" int64_t {funcname}(void * data) {{
    int64_t res;

    {varname} = ({dtype} *) data;

    res = 0;

    return res;
}}
"""

t_varmap_scalar = """
extern "C" int64_t {funcname}(void * data) {{
    int64_t res;

    {varname} = *({dtype} *) data;

    res = 0;

    return res;
}}
"""


class CppKernel(KernelBase):

    name = "cpp"
    lang = "cpp"

    # dtype: ( C type name, ctype )
    # TODO: dynamically generates on loading
#    dtypemap = {
#        "int32": ["int32_t", c_int32],
#        "int64": ["int64_t", c_int64],
#        "float32": ["float", c_float],
#        "float64": ["double", c_double]
#    }

    def gen_code(self, compiler):

        macros = {}
        includes = self.add_includes()

        main_fmt = {
            "preprop": self.gen_preprop(),
            "kernel":self.gen_kernel(),
            "varmap":self._gen_varmap(),
            "startmain":self.gen_startmain(),
            "include":self.get_include(),
        }
        main = t_main.format(**main_fmt)

        #print(main)
        #import pdb; pdb.set_trace()

        return includes + [main], macros

    def get_dtype(self, arg):
        return c_dtypemap[arg["data"].dtype.name][0]

    def gen_preprop(self):

        typedefs = []
        consts = []
        macros = []

        # TYPE(x), SHAPE(x, 0), SIZE(x), ARG(x), DVAR(x), FLATTEN(x)

        macros.append("#define TYPE(varname) apy_type_##varname")
        macros.append("#define SHAPE(varname, dim) apy_shape_##varname##dim")
        macros.append("#define SIZE(varname) apy_size_##varname")
        macros.append("#define ARG(varname) apy_type_##varname varname apy_shapestr_##varname")
        macros.append("#define VAR(varname) (*apy_ptr_##varname)")
        macros.append("#define DVAR(varname) (*apy_dptr_##varname)")
        macros.append("#define PTR(varname) apy_ptr_##varname")
        macros.append("#define DPTR(varname) apy_dptr_##varname")
        macros.append("#define FLATTEN(varname) accelpy_var_##varname")

        for arg in self.data:
            dtype = self.get_dtype(arg)
            name = arg["curname"]
            ndim = arg["data"].ndim

            consts.append("const int apy_size_%s = %d;" % (name, arg["data"].size))
            typedefs.append("typedef %s apy_type_%s;" % (dtype, name))

            if ndim > 0:

                shapestr = "".join(["[%d]"%s for s in arg["data"].shape])
                macros.append("#define apy_shapestr_%s %s" % (name, shapestr))
                for d, s in enumerate(arg["data"].shape):
                    consts.append("const int apy_shape_%s%d = %d;" % (name, d, s))
            else:
                pass

        return "\n".join(macros) + "\n\n" + "\n".join(typedefs) + "\n\n" + "\n".join(consts)

    def _gen_varmap(self):

        out = []

        for arg in self.data:
            dtype = self.get_dtype(arg)
            funcname = getname_varmap(arg)
            varname = "accelpy_var_" + arg["curname"]
            ndim = arg["data"].ndim

            if ndim > 0:
                out.append("%s * %s;" % (dtype, varname))
                out.append(t_varmap.format(funcname=funcname, varname=varname,
                            dtype=dtype))
            else:
                out.append("%s %s;" % (dtype, varname))
                out.append(t_varmap_scalar.format(funcname=funcname,
                            varname=varname, dtype=dtype))

        return "\n".join(out)

    def gen_kernel(self):

        args = []
        shapes = []

        for arg in self.data:
            ndim = arg["data"].ndim
            dtype = self.get_dtype(arg)
            name = arg["curname"]

            if ndim > 0:
                shape0 = "".join(["[%d]"%s for s in arg["data"].shape])
                shape1 = ",".join([str(s) for s in arg["data"].shape])

                shapes.append("int shape_%s[%d] = {%s};" % (name, ndim, shape1))
                args.append("%s %s%s" % (dtype, name, shape0))

            else:
                args.append("%s %s" % (dtype, name))

        return t_kernel.format(spec="\n".join(self.section.body),
                kernelargs=", ".join(args), shape="\n".join(shapes))

    def gen_startmain(self):

        argdefs = []
        startargs = []

        fmt = "{dtype}(*apy_ptr_{name}){shape} = reinterpret_cast<{dtype}(*){shape}>(accelpy_var_{name});"

        for arg in self.data:
            dtype = self.get_dtype(arg)
            name = arg["curname"]
            ndim = arg["data"].ndim

            if ndim > 0:
                shape = "".join(["[%d]"%s for s in arg["data"].shape])
                argdefs.append(fmt.format(dtype=dtype, name=name, shape=shape))
                argdefs.append("%s(*apy_dptr_%s)%s;" % (dtype, name, shape))
                startargs.append("(*apy_ptr_" + name + ")")

            else:
                startargs.append("accelpy_var_" + name)

        return "\n".join(argdefs) + "\n\n" + "res = accelpy_kernel(%s);" % ", ".join(startargs)


class OpenmpCppKernel(CppKernel):
    name = "openmp_cpp"


class OpenaccCppKernel(CppKernel):
    name = "openacc_cpp"


class OmptargetCppKernel(CppKernel):
    name = "omptarget_cpp"

OmptargetCppKernel.avails[OmptargetCppKernel.name] = OmptargetCppKernel
OpenaccCppKernel.avails[OpenaccCppKernel.name] = OpenaccCppKernel
OpenmpCppKernel.avails[OpenmpCppKernel.name] = OpenmpCppKernel
CppKernel.avails[CppKernel.name] = CppKernel


##########################
#  AccelData Code templates
##########################

t_accdata = """
extern "C" {{

{vardecls}

int dataenter({enterargs})
{{
    {enterassigns}

    {enterdir} {entermaps}

    return 0;
}}

int dataexit()
{{
    {exitdir} {exitmaps}

    return 0;
}}

}}
"""

class CppAccelData(AccelDataBase):

    name = "cpp"
    lang = "cpp"

    def get_dtype(self, arg):
        return c_dtypemap[arg["data"].dtype.name][0]

    def gen_code(self, compiler):
        pass

class OpenmpCppAccelData(CppAccelData):
    name = "openmp_cpp"


class CppOpenAccelData(CppAccelData):

    offload = True
    
    @abc.abstractmethod
    def clause_mapto(self, mapto):
        pass

    @abc.abstractmethod
    def clause_maptofrom(self, maptofrom):
        pass

    @abc.abstractmethod
    def clause_mapalloc(self, mapalloc):
        pass

    @abc.abstractmethod
    def clause_mapfrom(self, mapfrom):
        pass

    @abc.abstractmethod
    def enterdirect(self):
        pass

    @abc.abstractmethod
    def exitdirect(self):
        pass

    def gen_code(self, compiler):

        #import pdb; pdb.set_trace()
        macros = {}

        vardecls = []
        enterargs = []
        enterassigns = []
        mapto = []
        maptofrom = []
        mapalloc = []
        mapfrom = []

        for item in self.mapto:
            ndim = item["data"].ndim
            dtype = self.get_dtype(item)
            lname = "apy_acceldata_lvar_" + str(item["index"])
            gname = "apy_acceldata_gvar_" + str(item["index"])

            item["modname"] = gname

            enterargs.append("%s * %s" % (dtype, lname))
            vardecls.append("%s * %s;" % (dtype, gname))
            enterassigns.append("%s = %s;" % (gname, lname))
            mapto.append("%s[0:%d]" % (gname, item["data"].size))

        for item in self.maptofrom:
            ndim = item["data"].ndim
            dtype = self.get_dtype(item)
            lname = "apy_acceldata_lvar_" + str(item["index"])
            gname = "apy_acceldata_gvar_" + str(item["index"])

            item["modname"] = gname

            enterargs.append("%s * %s" % (dtype, lname))
            vardecls.append("%s * %s;" % (dtype, gname))
            enterassigns.append("%s = %s;" % (gname, lname))
            maptofrom.append("%s[0:%d]" % (gname, item["data"].size))

        for item in self.mapalloc:
            ndim = item["data"].ndim
            dtype = self.get_dtype(item)
            lname = "l" + str(item["index"])
            gname = "g" + str(item["index"])

            item["modname"] = gname

            enterargs.append("%s * %s" % (dtype, lname))
            vardecls.append("%s * %s;" % (dtype, gname))
            enterassigns.append("%s = %s;" % (gname, lname))
            mapalloc.append("%s[0:%d]" % (gname, item["data"].size))

        for item in self.mapfrom:
            ndim = item["data"].ndim
            dtype = self.get_dtype(item)
            lname = "apy_acceldata_lvar_" + str(item["index"])
            gname = "apy_acceldata_gvar_" + str(item["index"])

            item["modname"] = gname

            enterargs.append("%s * %s" % (dtype, lname))
            vardecls.append("%s * %s;" % (dtype, gname))
            enterassigns.append("%s = %s;" % (gname, lname))
            mapfrom.append("%s[0:%d]" % (gname, item["data"].size))
            mapalloc.append("%s[0:%d]" % (gname, item["data"].size))

        entermaps = "%s %s %s" % (self.clause_mapto(mapto),
                        self.clause_maptofrom(maptofrom, True),
                        self.clause_mapalloc(mapalloc))
        exitmaps = "%s %s" % (self.clause_maptofrom(maptofrom, False),
                                self.clause_mapfrom(mapfrom))

        enterdir = self.enterdirect()
        exitdir = self.exitdirect()

        acceldata_fmt = {
            "vardecls": "\n".join(vardecls),
            "enterargs": ", ".join(enterargs),
            "enterassigns": "\n".join(enterassigns),
            "enterdir": enterdir,
            "entermaps": entermaps,
            "exitdir": exitdir,
            "exitmaps": exitmaps,
        }

        acceldata = t_accdata.format(**acceldata_fmt)

        #print(acceldata)
        #import pdb; pdb.set_trace()

        return [acceldata], macros


class OpenaccCppAccelData(CppOpenAccelData):
    name = "openacc_cpp"

    def clause_mapto(self, mapto):
        return "copyin(%s)" % ", ".join(mapto) if mapto else ""

    def clause_maptofrom(self, maptofrom, isenter):

        if isenter:
            return "copyin(%s)" % ", ".join(maptofrom) if maptofrom else ""

        else:
            return "copyout(%s)" % ", ".join(maptofrom) if maptofrom else ""

    def clause_mapalloc(self, mapalloc):
        return "create(%s)" % ", ".join(mapalloc) if mapalloc else ""

    def clause_mapfrom(self, mapfrom):
        return "copyout(%s)" % ", ".join(mapfrom) if mapfrom else ""

    def enterdirect(self):
        return "#pragma acc enter data"

    def exitdirect(self):
        return "#pragma acc exit data"

class OmptargetCppAccelData(CppOpenAccelData):
    name = "omptarget_cpp"

    def clause_mapto(self, mapto):
        return "map(to: %s)" % ", ".join(mapto) if mapto else ""

    def clause_maptofrom(self, maptofrom, isenter):
        if isenter:
            return ("map(tofrom: %s)" % ", ".join(maptofrom)
                    if maptofrom else "")
        else:
            return ""

    def clause_mapalloc(self, mapalloc):
        return "map(alloc: %s)" % ", ".join(mapalloc) if mapalloc else ""

    def clause_mapfrom(self, mapfrom):
        return "map(from: %s)" % ", ".join(mapfrom) if mapfrom else ""

    def enterdirect(self):
        return "#pragma omp target enter data"

    def exitdirect(self):
        return "#pragma omp target exit data"


AccelDataBase.avails[OmptargetCppAccelData.name] = OmptargetCppAccelData
AccelDataBase.avails[OpenaccCppAccelData.name] = OpenaccCppAccelData
AccelDataBase.avails[OpenmpCppAccelData.name] = OpenmpCppAccelData
AccelDataBase.avails[CppAccelData.name] = CppAccelData



"Constant module"

name            = "accelpy"
version         = "0.5.2"
description     = "Scalable Accelerator Interface in Python"
long_description= "Scalable Accelerator Interface in Python"
author          = "Youngsung Kim"
vendors         = ("cray", "amd", "nvidia", "intel", "pgi", "ibm", "gnu")

NOCACHE         = 0
FILECACHE       = 1
MEMCACHE        = 2 

NODEBUG         = 0
MINDEBUG        = 1
MAXDEBUG        = 2

NOPROF          = 0
MINPROF         = 1
MAXPROF         = 2

# priorities
lang_priority = ("fortran", "cpp")
accel_priority = ("omptarget_fortran", "omptarget_cpp", "openmp_fortran",
                    "openmp_cpp", "openacc_fortran", "openacc_cpp", "hip",
                    "cuda", "fortran", "cpp")
vendor_priority = ("cray", "amd", "nvidia", "intel", "pgi", "ibm", "gnu")




"""accelpy Fortran Accelerator module"""

import abc

from uuid import uuid4
from accelpy.accel import AccelDataBase
from accelpy.kernel import KernelBase
from accelpy.util import fortline_pack, f_dtypemap, getname_varmap
from ctypes import c_int32, c_int64, c_float, c_double


##########################
#  Kernel Code templates
##########################

t_module = """
MODULE {modname}
    USE, INTRINSIC :: ISO_C_BINDING
    IMPLICIT NONE

{datavars}

END MODULE
"""

t_main = """
{include}

{varmap}

INTEGER (C_INT64_T) FUNCTION accelpy_start() BIND(C, name="accelpy_start")

    USE, INTRINSIC :: ISO_C_BINDING
    USE {modname}, ONLY : {usevarnames}
    IMPLICIT NONE

    accelpy_start = accelpy_kernel({usevarnames})

CONTAINS

{kernel}

END FUNCTION

INTEGER (C_INT64_T) FUNCTION accelpy_stop() BIND(C, name="accelpy_stop")
    USE, INTRINSIC :: ISO_C_BINDING
    USE {modname}, ONLY : {usevarnames}
    IMPLICIT NONE

    accelpy_stop = 0

END FUNCTION
"""

t_kernel = """
INTEGER (C_INT64_T) FUNCTION accelpy_kernel({varnames})
    USE, INTRINSIC :: ISO_C_BINDING
    IMPLICIT NONE

    {typedecls}

    {spec}

    accelpy_kernel = 0

END FUNCTION
"""

t_varmap = """
INTEGER (C_INT64_T) FUNCTION {funcname} (data) BIND(C, name="{funcname}")
    USE, INTRINSIC :: ISO_C_BINDING
    USE {modname}, ONLY : {varname}
    IMPLICIT NONE

    {dtype}, DIMENSION({bound}), INTENT(IN), TARGET :: data

    {varname} => data

    {funcname} = 0

END FUNCTION
"""

t_varmap_scalar = """
INTEGER (C_INT64_T) FUNCTION {funcname} (data) BIND(C, name="{funcname}")
    USE, INTRINSIC :: ISO_C_BINDING
    USE {modname}, ONLY : {varname}
    IMPLICIT NONE

    {dtype}, DIMENSION(1), INTENT(IN) :: data

    {varname} = data(1)

    {funcname} = 0

END FUNCTION
"""

class FortranKernel(KernelBase):

    name = "fortran"
    lang = "fortran"

#    dtypemap = {
#        "int32": ["INTEGER (C_INT32_T)", c_int32],
#        "int64": ["INTEGER (C_INT64_T)", c_int64],
#        "float32": ["REAL (C_FLOAT)", c_float],
#        "float64": ["REAL (C_DOUBLE)", c_double]
#    }

    def gen_code(self, compiler):

        macros = {}
        includes = self.add_includes()

        modname  = "mod_" + uuid4().hex[:10]

        module_fmt = {
            "modname": modname,
            "datavars": self._get_datavars(),
        }
        module = t_module.format(**module_fmt)

        main_fmt = {
            "modname":modname,
            "varmap":self._gen_varmap(modname),
            "kernel":self._gen_kernel(),
            "usevarnames":self._gen_usevars(),
            "include":self.get_include(),
        }
        main = t_main.format(**main_fmt)

        #print(module)
        #print(main)
        #import pdb; pdb.set_trace()

        return includes + [module, main], macros

    def get_dtype(self, arg):
        return f_dtypemap[arg["data"].dtype.name][0]

    def _get_datavars(self):

        out = []

        for arg in self.data:
            ndim = arg["data"].ndim
            dtype = self.get_dtype(arg)
            varname = arg["curname"]

            if ndim > 0:
                bound = ",".join([":"]*ndim)
                out.append("%s, DIMENSION(%s), POINTER :: %s" % (dtype, bound, varname))

            else:
                out.append("%s :: %s" % (dtype, varname))

        return "\n".join(out)

    def _gen_varmap(self, modname):

        out = []

        for arg in self.data:
            dtype = self.get_dtype(arg)
            funcname = getname_varmap(arg)
            ndim = arg["data"].ndim

            if ndim > 0:
                bound = ",".join([str(s) for s in arg["data"].shape])
                out.append(t_varmap.format(funcname=funcname, varname=arg["curname"],
                            modname=modname, bound=bound, dtype=dtype))
            else:
                out.append(t_varmap_scalar.format(funcname=funcname, modname=modname,
                            varname=arg["curname"], dtype=dtype))

        return "\n".join(out)

    def _gen_kernel(self):

        _names = []
        typedecls = []

        for arg in self.data:
            ndim = arg["data"].ndim
            dtype = self.get_dtype(arg)
            varname = arg["curname"]

            _names.append(varname)

            if ndim > 0:
                if ("attrspec" in self.section.kwargs and
                    varname in self.section.kwargs["attrspec"] and
                    "dimension" in self.section.kwargs["attrspec"][varname]):
                    bound = self.section.kwargs["attrspec"][varname]["dimension"]

                else:
                    bound = ",".join([":"]*ndim)


                typedecls.append("%s, DIMENSION(%s), INTENT(INOUT) :: %s" % (
                            dtype, bound, varname))

            else:
                typedecls.append("%s, INTENT(IN) :: %s" % (dtype, varname))

        names = fortline_pack(_names)

        return t_kernel.format(spec="\n".join(self.section.body),
                typedecls="\n".join(typedecls), varnames="\n".join(names))

    def _gen_usevars(self):

        names = []

        for arg in self.data:

            names.append(arg["curname"])

        lines = [""]
        maxlen = 72

        for name in names:
            if len(lines[-1]) + len(name) > maxlen:

                lines[-1] += " &"
                lines.append("        &, %s" % name)

            elif lines[-1] == "":
                lines[-1] += name

            else:
                lines[-1] += ", " + name

        return "\n".join(lines)

class OpenmpFortranKernel(FortranKernel):
    name = "openmp_fortran"

class OpenaccFortranKernel(FortranKernel):
    name = "openacc_fortran"

class OmptargetFortranKernel(OpenmpFortranKernel):
    name = "omptarget_fortran"

OmptargetFortranKernel.avails[OmptargetFortranKernel.name] = OmptargetFortranKernel
OpenaccFortranKernel.avails[OpenaccFortranKernel.name] = OpenaccFortranKernel
OpenmpFortranKernel.avails[OpenmpFortranKernel.name] = OpenmpFortranKernel
FortranKernel.avails[FortranKernel.name] = FortranKernel



##########################
#  AccelData Code templates
##########################

t_accdata = """
MODULE {modname}
    USE, INTRINSIC :: ISO_C_BINDING
    IMPLICIT NONE

{modvardecls}

PUBLIC {modpublics}

CONTAINS

INTEGER (C_INT64_T) FUNCTION dataenter({enterargs}) BIND(C, name="dataenter")
    USE, INTRINSIC :: ISO_C_BINDING

    {entertypedecls}

    {enterassigns}

    {enterdir} {entermaps}

    dataenter = 0

END FUNCTION

INTEGER (C_INT64_T) FUNCTION dataexit() BIND(C, name="dataexit")
    USE, INTRINSIC :: ISO_C_BINDING

    {exitdir} {exitmaps}

    dataexit = 0

END FUNCTION

END MODULE
"""

class FortranAccelData(AccelDataBase):

    name = "fortran"
    lang = "fortran"

#    dtypemap = {
#        "int32": ["INTEGER (C_INT32_T)", c_int32],
#        "int64": ["INTEGER (C_INT64_T)", c_int64],
#        "float32": ["REAL (C_FLOAT)", c_float],
#        "float64": ["REAL (C_DOUBLE)", c_double]
#    }

    def get_dtype(self, arg):
        return f_dtypemap[arg["data"].dtype.name][0]

    def gen_code(self, compiler):
        pass

class OpenmpFortranAccelData(FortranAccelData):
    name = "openmp_fortran"


class FortranOpenAccelData(FortranAccelData):

    offload = True
    
    @abc.abstractmethod
    def clause_mapto(self, mapto):
        pass

    @abc.abstractmethod
    def clause_maptofrom(self, maptofrom):
        pass

    @abc.abstractmethod
    def clause_mapalloc(self, mapalloc):
        pass

    @abc.abstractmethod
    def clause_mapfrom(self, mapfrom):
        pass

    @abc.abstractmethod
    def enterdirect(self):
        pass

    @abc.abstractmethod
    def exitdirect(self):
        pass

    def gen_code(self, compiler):

        macros = {}

        modname  = "mod_" + uuid4().hex[:10]

        modvardecls = []
        modpublics = []
        enterargs = []
        entertypedecls = []
        enterassigns = []
        mapto = []
        maptofrom = []
        mapalloc = []
        mapfrom = []

        for item in self.mapto:
            ndim = item["data"].ndim
            dtype = self.get_dtype(item)
            lname = "apy_acceldata_lvar_" + str(item["index"])
            gname = "apy_acceldata_gvar_" + str(item["index"])

            item["modname"] = gname

            enterargs.append(lname)
            shape = ",".join([":"] * ndim)
            mapto.append("%s(%s)" % (gname, shape))
            bound = ",".join([str(s) for s in item["data"].shape])
            entertypedecls.append("%s, DIMENSION(%s), INTENT(INOUT), TARGET :: %s" % (
                            dtype, bound, lname))

            modpublics.append(gname)
            bound = ",".join([":"]*ndim)
            modvardecls.append("%s, DIMENSION(%s), POINTER :: %s" % (
                            dtype, bound, gname))

            enterassigns.append("%s => %s" % (gname, lname))

        for item in self.maptofrom:
            ndim = item["data"].ndim
            dtype = self.get_dtype(item)
            lname = "apy_acceldata_lvar_" + str(item["index"])
            gname = "apy_acceldata_gvar_" + str(item["index"])

            item["modname"] = gname

            enterargs.append(lname)
            shape = ",".join([":"] * ndim)
            maptofrom.append("%s(%s)" % (gname, shape))
            bound = ",".join([str(s) for s in item["data"].shape])
            entertypedecls.append("%s, DIMENSION(%s), INTENT(INOUT), TARGET :: %s" % (
                            dtype, bound, lname))

            modpublics.append(gname)
            bound = ",".join([":"]*ndim)
            modvardecls.append("%s, DIMENSION(%s), POINTER :: %s" % (
                            dtype, bound, gname))

            enterassigns.append("%s => %s" % (gname, lname))

        for item in self.mapalloc:
            ndim = item["data"].ndim
            dtype = self.get_dtype(item)
            lname = "l" + str(item["index"])
            gname = "g" + str(item["index"])

            item["modname"] = gname

            enterargs.append(lname)
            shape = ",".join([":"] * ndim)
            mapalloc.append("%s(%s)" % (gname, shape))
            bound = ",".join([str(s) for s in item["data"].shape])
            entertypedecls.append("%s, DIMENSION(%s), INTENT(INOUT), TARGET :: %s" % (
                            dtype, bound, lname))

            modpublics.append(gname)
            bound = ",".join([":"]*ndim)
            modvardecls.append("%s, DIMENSION(%s), POINTER :: %s" % (
                            dtype, bound, gname))

            enterassigns.append("%s => %s" % (gname, lname))

        for item in self.mapfrom:
            ndim = item["data"].ndim
            dtype = self.get_dtype(item)
            lname = "apy_acceldata_lvar_" + str(item["index"])
            gname = "apy_acceldata_gvar_" + str(item["index"])

            item["modname"] = gname

            enterargs.append(lname)
            shape = ",".join([":"] * ndim)
            mapfrom.append("%s(%s)" % (gname, shape))
            mapalloc.append("%s(%s)" % (gname, shape))
            bound = ",".join([str(s) for s in item["data"].shape])
            entertypedecls.append("%s, DIMENSION(%s), INTENT(INOUT), TARGET :: %s" % (
                            dtype, bound, lname))

            modpublics.append(gname)
            bound = ",".join([":"]*ndim)
            modvardecls.append("%s, DIMENSION(%s), POINTER :: %s" % (
                            dtype, bound, gname))

            enterassigns.append("%s => %s" % (gname, lname))

        entermaps = "%s %s %s" % (self.clause_mapto(mapto),
                        self.clause_maptofrom(maptofrom, True),
                        self.clause_mapalloc(mapalloc))
        exitmaps = "%s %s" % (self.clause_maptofrom(maptofrom, False),
                                self.clause_mapfrom(mapfrom))

        enterdir = self.enterdirect()
        exitdir = self.exitdirect()

        acceldata_fmt = {
            "modname": modname,
            "modvardecls": "\n".join(modvardecls),
            "modpublics": ", ".join(modpublics),
            "enterargs": ", ".join(enterargs),
            "entertypedecls": "\n".join(entertypedecls),
            "enterassigns": "\n".join(enterassigns),
            "enterdir": enterdir,
            "entermaps": entermaps,
            "exitdir": exitdir,
            "exitmaps": exitmaps,
        }

        acceldata = t_accdata.format(**acceldata_fmt)

        #print(acceldata)
        #import pdb; pdb.set_trace()

        return [acceldata], macros


class OpenaccFortranAccelData(FortranOpenAccelData):
    name = "openacc_fortran"

    def clause_mapto(self, mapto):
        return "copyin(%s)" % ", ".join(mapto) if mapto else ""

    def clause_maptofrom(self, maptofrom, isenter):

        if isenter:
            return "copyin(%s)" % ", ".join(maptofrom) if maptofrom else ""

        else:
            return "copyout(%s)" % ", ".join(maptofrom) if maptofrom else ""

    def clause_mapalloc(self, mapalloc):
        return "create(%s)" % ", ".join(mapalloc) if mapalloc else ""

    def clause_mapfrom(self, mapfrom):
        return "copyout(%s)" % ", ".join(mapfrom) if mapfrom else ""

    def enterdirect(self):
        return "!$acc enter data"

    def exitdirect(self):
        return "!$acc exit data"

class OmptargetFortranAccelData(FortranOpenAccelData):
    name = "omptarget_fortran"

    def clause_mapto(self, mapto):
        return "map(to: %s)" % ", ".join(mapto) if mapto else ""

    def clause_maptofrom(self, maptofrom, isenter):
        if isenter:
            return ("map(tofrom: %s)" % ", ".join(maptofrom)
                    if maptofrom else "")
        else:
            return ""

    def clause_mapalloc(self, mapalloc):
        return "map(alloc: %s)" % ", ".join(mapalloc) if mapalloc else ""

    def clause_mapfrom(self, mapfrom):
        return "map(from: %s)" % ", ".join(mapfrom) if mapfrom else ""

    def enterdirect(self):
        return "!$omp target enter data"

    def exitdirect(self):
        return "!$omp target exit data"


AccelDataBase.avails[OmptargetFortranAccelData.name] = OmptargetFortranAccelData
AccelDataBase.avails[OpenaccFortranAccelData.name] = OpenaccFortranAccelData
AccelDataBase.avails[OpenmpFortranAccelData.name] = OpenmpFortranAccelData
AccelDataBase.avails[FortranAccelData.name] = FortranAccelData



"accelpy setup module."

import os, json
from setuptools import setup, find_packages
from setuptools.command.develop import develop
from setuptools.command.install import install
from distutils.errors import DistutilsClassError

bdist_wheel = None

try:
    import wheel.bdist_wheel
    class bdist_wheel(wheel.bdist_wheel.bdist_wheel):
        def run(self, *args, **kwargs):
            pass
            #raise DistutilsClassError("No!")
except ModuleNotFoundError:
    pass

genv = {'__builtins__' : None}
consts = {}

with open(os.path.join("accelpy", "const.py")) as fp:
    exec(fp.read(), genv, consts)

def _setcfg():

    cfgdir = os.path.join(os.path.expanduser("~"), ".accelpy")

    libdir = os.path.join(cfgdir, "lib")
    cfgfile = os.path.join(cfgdir, "config")

    for vendor in consts["vendors"]:
        vendor_path = os.path.join(libdir, vendor)
        if not os.path.isdir(vendor_path):
            try:
                os.makedirs(vendor_path)

            except FileExistsError:
                pass

    config = {
        "libdir": libdir,
        "blddir": "",
    }

    if not os.path.isfile(cfgfile):
        with open(cfgfile, "w")  as f:
            json.dump(config, f)


class DevelopCommand(develop):
    def run(self):
        _setcfg()
        develop.run(self)


class InstallCommand(install):
    def run(self):
        _setcfg()
        install.run(self)


def main():

    install_requires = ["numpy"]
    console_scripts = ["accelpy=accelpy.command:main"]
    keywords = ["GPU", "CPU", "Accelerator", "Cuda", "Hip",
                "OpenAcc", "OpenMP", "Numpy", "C++", "Fortran", "accelpy"]

    setup(
        name=consts["name"],
        version=consts["version"],
        description=consts["description"],
        long_description=consts["long_description"],
        author=consts["author"],
        author_email="youngsung.kim.act2@gmail.com",
        classifiers=[
            "Development Status :: 3 - Alpha",
            "Intended Audience :: Science/Research",
            "Topic :: Scientific/Engineering",
            "License :: OSI Approved :: MIT License",
            "Programming Language :: Python :: 3",
            "Programming Language :: Python :: 3.5",
            "Programming Language :: Python :: 3.6",
            "Programming Language :: Python :: 3.7",
            "Programming Language :: Python :: 3.8",
        ],
        cmdclass={
            'develop': DevelopCommand,
            'install': InstallCommand,
            'bdist_wheel': bdist_wheel
        },
        keywords=keywords,
        packages=find_packages(exclude=["tests"]),
        include_package_data=True,
        install_requires=install_requires,
        entry_points={ "console_scripts": console_scripts },
        project_urls={
            "Bug Reports": "https://github.com/grnydawn/accelpy/issues",
            "Source": "https://github.com/grnydawn/accelpy",
        }
    )

if __name__ == '__main__':
    import multiprocessing
    multiprocessing.freeze_support()
    main()
